import importlib
from unittest import TestCase
from unittest.mock import patch

from servey.action.action import action, get_action


class TestServeyDirect(TestCase):
    def test_servey_direct(self):
        was_run = False

        @action
        def my_action() -> str:
            nonlocal was_run
            was_run = True
            return "Action was run"

        with (
            patch("sys.argv", ["servey", "--run=action", "--action=my_action"]),
            patch(
                "servey.finder.action_finder_abc.find_actions",
                return_value=[get_action(my_action)],
            ),
        ):
            _reload_module()
        self.assertTrue(was_run)

    def test_servey_direct_no_action(self):
        @action
        def my_action() -> str:
            """Dummy"""

        with (
            patch("sys.argv", ["servey", "--run=action", "--action=does_not_exist"]),
            patch(
                "servey.finder.action_finder_abc.find_actions",
                return_value=[get_action(my_action)],
            ),
        ):
            with self.assertRaises(ValueError):
                _reload_module()


_module = None


def _reload_module():
    global _module
    if _module:
        importlib.reload(_module)
    else:
        from servey.servey_direct import __main__

        _module = __main__



import os
from dataclasses import dataclass, field
from logging import getLogger
from typing import Iterator

from starlette.routing import Route, WebSocketRoute, Mount
from starlette.staticfiles import StaticFiles

from servey.servey_starlette.route_factory.route_factory_abc import (
    RouteFactoryABC,
)

LOGGER = getLogger(__name__)


@dataclass
class StrawberryStarletteRouteFactory(RouteFactoryABC):
    """Route factory which adds graphql"""

    graphql_path: str = field(
        default_factory=lambda: os.environ.get("SERVEY_GRAPHQL_PATH") or "/graphql"
    )
    debug: bool = field(
        default_factory=lambda: int(os.environ.get("SERVER_DEBUG", "1")) == 1
    )

    def create_routes(self) -> Iterator[Route]:
        # Create an authenticator object based on username and password
        try:
            from servey.servey_strawberry.schema_factory import (
                create_schema,
            )
            from strawberry.asgi import GraphQL

            schema = create_schema()
            if not schema:
                return
            graphql_app = GraphQL(schema, debug=self.debug)
            yield Route(path=self.graphql_path, methods=["post"], endpoint=graphql_app)
            yield WebSocketRoute(path=self.graphql_path, endpoint=graphql_app)
            if self.debug:
                yield Mount(
                    "/graphiql",
                    app=StaticFiles(
                        packages=["servey.servey_strawberry"],
                        html=True,
                    ),
                    name="graphiql",
                )

        except ModuleNotFoundError:
            LOGGER.error("Graphql not found - skipping")



from unittest import TestCase

from servey.security.authorizer.authorizer_factory_abc import (
    create_authorizer,
    get_default_authorizer,
)
from servey.security.authorizer.jwt_authorizer import JwtAuthorizer


class TestAuthorizerFactory(TestCase):
    def test_create_authorizer(self):
        authorizer1 = create_authorizer()
        self.assertIsInstance(authorizer1, JwtAuthorizer)
        authorizer2 = create_authorizer()
        self.assertIsInstance(authorizer2, JwtAuthorizer)
        self.assertNotEqual(authorizer1.private_key, authorizer2.private_key)

    def test_get_default_authorizer(self):
        self.assertIs(get_default_authorizer(), get_default_authorizer())



from unittest import TestCase

from servey.action.batch_invoker import noop


class TestBatchInvoker(TestCase):
    def test_noop(self):
        item = "foobar"
        self.assertEqual([item], noop(item))



from datetime import datetime
from unittest import TestCase
from unittest.mock import patch

from servey.action.action import get_action, action
from servey.servey_thread.fixed_rate_trigger_thread import FixedRateTriggerThread
from servey.trigger.fixed_rate_trigger import FixedRateTrigger


class TestThreadedApp(TestCase):
    def test_threaded(self):
        print_time()
        print_time_action = get_action(print_time)
        with (
            patch(
                "servey.finder.action_finder_abc.find_actions",
                return_value=[print_time_action],
            ),
            patch(
                "servey.servey_thread.fixed_rate_trigger_thread.FixedRateTriggerThread.start",
                return_value=None,
            ),
        ):
            import servey.servey_thread.__main__ as thread_main

            self.assertEqual(1, len(thread_main._THREADS))

    def test_run(self):
        def raise_error():
            assert False

        def do_not_sleep(_: int):
            """Speed up iteration"""

        thread = FixedRateTriggerThread(raise_error, FixedRateTrigger(1), False)
        with patch("time.sleep", do_not_sleep):
            with self.assertRaises(AssertionError):
                thread.run()


# noinspection PyTypeChecker
@action(triggers=(FixedRateTrigger(10),))
def print_time() -> type(None):
    print(f"Time is : {datetime.now()}")



from asyncio import get_event_loop
from dataclasses import dataclass
from typing import Optional, List

from servey.servey_aws import is_lambda_env
from servey.servey_celery import has_celery_broker
from servey.subscription.subscription import Subscription, T
from servey.subscription.subscription_service import (
    SubscriptionServiceABC,
    SubscriptionServiceFactoryABC,
)


@dataclass
class AsyncioSubscriptionService(SubscriptionServiceABC):
    """Subscription service backed by asyncio"""

    def publish(
        self,
        subscription: Subscription[T],
        event: T,
    ):
        loop = get_event_loop()
        for action in subscription.action_subscribers:
            loop.call_soon_threadsafe(action.fn, event)


class AsyncioSubscriptionServiceFactory(SubscriptionServiceFactoryABC):
    def create(
        self, subscriptions: List[Subscription]
    ) -> Optional[SubscriptionServiceABC]:
        if is_lambda_env() or has_celery_broker():
            # lambda environment, uses SQS instead, and if celery is
            # present we should use it
            return
        has_subscribers_with_actions = next(
            (True for s in subscriptions if s.action_subscribers), False
        )
        if has_subscribers_with_actions:
            return AsyncioSubscriptionService()



from unittest import TestCase

from marshy import dump, load

from servey.util.singleton_abc import SingletonABC


class MySingleton(SingletonABC):
    pass


class TestSingleton(TestCase):
    def test_singleton(self):
        self.assertIs(MySingleton(), MySingleton())

    def test_eq(self):
        self.assertEqual(MySingleton(), MySingleton())

    def test_repr(self):
        self.assertEqual(str(MySingleton()), "MySingleton")

    def test_load_and_dump(self):
        dumped = dump(MySingleton())
        loaded = load(MySingleton, dumped)
        assert loaded is MySingleton()



from datetime import datetime

from servey.action.action import action
from servey.trigger.fixed_rate_trigger import FixedRateTrigger
from servey.trigger.web_trigger import WEB_GET
from tests.specs.number_spec.models import IntegerStats


@action(triggers=(WEB_GET,))
def integer_stats(value: int) -> IntegerStats:
    return IntegerStats(value)


@action
def integer_stats_consumer(stats: IntegerStats):
    print(f"IntegerStats: {stats.value}: Factorial: {stats.factorial}")


@action(triggers=(FixedRateTrigger(300),))
def integer_stats_publisher():
    integer_stats_ = IntegerStats(int(datetime.now().timestamp()) % 200)
    from tests.specs.number_spec.subscriptions import integer_stats_queue

    integer_stats_queue.publish(integer_stats_)



import os
from unittest import TestCase
from unittest.mock import patch
from uuid import UUID

from servey.action.action import action, get_action
from servey.subscription.subscription import subscription
from servey.trigger.fixed_rate_trigger import FixedRateTrigger


class TestCeleryApp(TestCase):
    def test_app(self):
        @action(triggers=(FixedRateTrigger(10),))
        def ping() -> UUID:
            """No implementation required"""

        # noinspection PyUnusedLocal
        @action
        def consume_message(message: str):
            """No implementation required"""

        subscription_ = subscription(
            str,
            name="message_broadcaster",
            action_subscribers=(get_action(consume_message),),
        )

        with (
            patch(
                "servey.finder.action_finder_abc.find_actions",
                return_value=[get_action(ping), get_action(consume_message)],
            ),
            patch(
                "servey.finder.subscription_finder_abc.find_subscriptions",
                return_value=[subscription_],
            ),
            patch.dict(os.environ, dict(CELERY_BROKER="redis://localhost:6379/0")),
            patch("celery.app.task.Task.apply_async", return_value=None),
        ):
            from servey.servey_celery.celery_app import app

            next(t for t in app.tasks if t.endswith("test_celery_app.ping"))
            next(t for t in app.tasks if t.endswith("test_celery_app.consume_message"))
            subscription_.publish("Some message")



import asyncio
import json
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Dict, Any
from unittest import TestCase
from email.utils import parsedate

from starlette.requests import Request
from starlette.responses import Response

from servey.action.action import action, Action, get_action
from servey.cache_control.timestamp_cache_control import TimestampCacheControl
from servey.cache_control.ttl_cache_control import TtlCacheControl
from servey.servey_starlette.action_endpoint.caching_action_endpoint import (
    CachingActionEndpoint,
)
from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory import (
    ActionEndpointFactory,
)
from servey.trigger.web_trigger import WEB_GET
from tests.servey_starlette.action_endpoint.test_action_endpoint import build_request


class TestCachingActionEndpoint(TestCase):
    def test_caching_input_get(self):
        @action(triggers=(WEB_GET,), cache_control=TtlCacheControl(30))
        def echo_get(val: str) -> str:
            return val

        action_endpoint = CachingActionEndpoint(
            ActionEndpointFactory().create(echo_get.__servey_action__, set(), [])
        )
        request = build_request(query_string="val=bar")
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(action_endpoint.execute(request))

        self.assertEqual(200, response.status_code)
        self.assertEqual("bar", json.loads(response.body))
        # noinspection SpellCheckingInspection
        self.assertEqual(
            "TCk/8BCnMPCXJ2EzHRtWeEeNQlwtxc79FtjyAFnkl/M=", response.headers["etag"]
        )
        self.assertEqual("private,max-age=29", response.headers["cache-control"])
        expected_expires = int(datetime.now().timestamp()) + 30
        expires = int(
            datetime(
                *(parsedate(response.headers["expires"])[:6]), tzinfo=timezone.utc
            ).timestamp()
        )
        self.assertAlmostEqual(expected_expires, expires, 1)

        # noinspection SpellCheckingInspection
        request = build_request(
            query_string="val=bar",
            headers={"If-Match": "TCk/8BCnMPCXJ2EzHRtWeEeNQlwtxc79FtjyAFnkl/M="},
        )
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(action_endpoint.execute(request))

        self.assertEqual(304, response.status_code)
        self.assertEqual(b"", response.body)

    def test_execute_with_context_500(self):
        expected_response = Response(b"", 500)
        # noinspection PyTypeChecker
        endpoint = CachingActionEndpoint(MockActionEndpoint(None, expected_response))
        loop = asyncio.get_event_loop()
        request = build_request()
        response = loop.run_until_complete(endpoint.execute_with_context(request, {}))
        self.assertEqual(500, response.status_code)
        self.assertEqual(b"", response.body)

    def test_execute_with_context_last_modified(self):
        expected_response = Response(
            json.dumps({"updated_at": "2022-01-01 00:00:00+00:00"}),
            200,
            {"Last-Modified": "Sat, 1 Jan 2022 00:00:00 GMT"},
        )
        # noinspection PyTypeChecker
        action_ = Action(None, "foobar", cache_control=TimestampCacheControl())
        # noinspection PyTypeChecker
        endpoint = CachingActionEndpoint(MockActionEndpoint(action_, expected_response))
        loop = asyncio.get_event_loop()
        request = build_request(
            headers={"If-Modified-Since": "Sun, 1 Feb 2022 00:00:00 GMT"}
        )
        response = loop.run_until_complete(endpoint.execute_with_context(request, {}))
        self.assertEqual(304, response.status_code)
        self.assertEqual(b"", response.body)

    def test_to_openapi_schema(self):
        # noinspection PyUnusedLocal
        @action(triggers=(WEB_GET,))
        def dummy_lookup(id: str) -> str:
            """No action required"""

        caching_action_endpoint = CachingActionEndpoint(
            action_endpoint=ActionEndpointFactory().create(
                get_action(dummy_lookup), set(), []
            )
        )
        schema = dict(paths={}, components={})
        caching_action_endpoint.to_openapi_schema(schema)
        expected_schema = {
            "components": {
                "ErrorResponse": {
                    "additionalProperties": False,
                    "description": "Content definition for error response",
                    "name": "ErrorResponse",
                    "properties": {
                        "error_code": {"type": "string"},
                        "error_msg": {
                            "anyOf": [{"type": "string"}, {"type": "null"}],
                            "default": None,
                        },
                    },
                    "required": ["error_code"],
                    "type": "object",
                }
            },
            "paths": {
                "/actions/dummy-lookup": {
                    "get": {
                        "operationId": "dummy_lookup",
                        "parameters": [
                            {
                                "in": "query",
                                "name": "id",
                                "required": True,
                                "schema": {"type": "string"},
                            }
                        ],
                        "responses": {
                            "200": {
                                "content": {
                                    "application/json": {"schema": {"type": "string"}}
                                },
                                "description": "Successful Response",
                            },
                            "304": {"content": {}, "description": "not_modified"},
                            "422": {"description": "Validation " "Error"},
                        },
                        "summary": "No action required",
                    }
                }
            },
        }
        self.assertEqual(expected_schema, schema)


@dataclass
class MockActionEndpoint:
    action: Action
    response: Response

    def get_action(self) -> Action:
        return self.action

    # noinspection PyUnusedLocal
    async def execute_with_context(
        self, request: Request, context: Dict[str, Any]
    ) -> Response:
        return self.response



from __future__ import annotations
from abc import abstractmethod, ABC
from typing import Dict, Any
from urllib.request import Request

from marshy.types import ExternalItemType
from requests import Response
from starlette.routing import Route

from servey.action.action import Action


class ActionEndpointABC(ABC):
    @abstractmethod
    def get_action(self) -> Action:
        """Get the action for this endpoint"""

    @abstractmethod
    def get_route(self) -> Route:
        """
        Convert this endpoint to a route which may be used by starlette
        """

    @abstractmethod
    async def execute_with_context(
        self, request: Request, context: Dict[str, Any]
    ) -> Response:
        """
        Execute the action for this endpoint with the request and overrides given
        """

    async def execute(self, request: Request) -> Response:
        response = await self.execute_with_context(request, {})
        return response

    @abstractmethod
    def to_openapi_schema(self, schema: ExternalItemType):
        """
        Merge the openapi schema for this endpoint into the schema given
        """



from unittest import TestCase

from marshy.marshaller import DatetimeMarshaller

from servey.cache_control.cache_header import CacheHeader
from servey.cache_control.timestamp_cache_control import TimestampCacheControl


class TestTimestampCacheControl(TestCase):
    def test_get_cache_header(self):
        cache_control = TimestampCacheControl(timestamp_marshaller=DatetimeMarshaller())
        cache_header = cache_control.get_cache_header_from_content(
            '{"values": [1, "str", true], "updated_at": "2022-12-01 02:03:04+00:00"}'.encode()
        )
        expected = CacheHeader()
        self.assertIsNotNone(expected, cache_header)



from dataclasses import dataclass
from typing import Optional

from servey.security.access_control.access_control_abc import (
    AccessControlABC,
)
from servey.security.authorization import Authorization


@dataclass(frozen=True)
class ScopeAccessControl(AccessControlABC):
    execute_scope: Optional[str]

    def is_executable(self, authorization: Optional[Authorization]) -> bool:
        return bool(authorization and self.execute_scope) and authorization.has_scope(
            self.execute_scope
        )



import os
from datetime import datetime
from unittest import TestCase
from unittest.mock import patch, MagicMock

from servey.security.authorization import ROOT, Authorization, AuthorizationError
from servey.servey_aws.authorizer.kms_authorizer_factory import KmsAuthorizerFactory


class TestKmsAuthorizer(TestCase):
    def test_authorizer(self):
        mock_client = MagicMock()
        mock_client.return_value.sign.return_value = {"Signature": b"NotARealSignature"}
        mock_client.return_value.get_public_key.return_value = {
            "PublicKey": b"NotARealPublicKey"
        }
        nbf = int(datetime.now().timestamp())
        exp = nbf + 3600
        with (
            patch.dict(os.environ, {"KMS_KEY_ID": "some_key"}),
            patch("boto3.client", mock_client),
            patch(
                "jwt.decode",
                return_value={
                    "sub": "some-sub",
                    "nbf": nbf,
                    "exp": exp,
                    "scope": "foo bar",
                },
            ),
        ):
            factory = KmsAuthorizerFactory()
            authorizer = factory.create_authorizer()
            token = authorizer.encode(ROOT)
            # noinspection PyUnresolvedReferences
            self.assertEqual("alias/some_key", authorizer.get_key_id(token))
            auth = authorizer.authorize(token)
            self.assertEqual(
                Authorization(
                    subject_id="some-sub",
                    not_before=datetime.fromtimestamp(nbf),
                    expire_at=datetime.fromtimestamp(exp),
                    scopes=frozenset(("foo", "bar")),
                ),
                auth,
            )

    def test_authorize_invalid(self):
        mock_client = MagicMock()
        with (
            patch.dict(os.environ, {"KMS_KEY_ID": "some_key"}),
            patch("boto3.client", mock_client),
        ):
            factory = KmsAuthorizerFactory()
            authorizer = factory.create_authorizer()
            with self.assertRaises(AuthorizationError):
                authorizer.authorize("notAValidToken")



import inspect
from dataclasses import field, dataclass
from typing import List, Optional, Set

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from schemey import get_default_schema_context, SchemaContext

from servey.action.action import Action
from servey.action.util import get_marshaller_for_params, get_schema_for_params
from servey.errors import ServeyError
from servey.servey_starlette.action_endpoint.action_endpoint import ActionEndpoint
from servey.servey_starlette.action_endpoint.action_endpoint_abc import (
    ActionEndpointABC,
)
from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory_abc import (
    ActionEndpointFactoryABC,
)
from servey.trigger.web_trigger import WebTrigger


@dataclass
class ActionEndpointFactory(ActionEndpointFactoryABC):
    priority: int = 100
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    schema_context: SchemaContext = field(default_factory=get_default_schema_context)
    validate_output: bool = True
    path_pattern: str = "/actions/{action_name}"

    def create(
        self,
        action: Action,
        skip_args: Set[str],
        factories: List[ActionEndpointFactoryABC],
    ) -> Optional[ActionEndpointABC]:
        web_triggers = [t for t in action.triggers if isinstance(t, WebTrigger)]
        if web_triggers:
            methods = {t.method for t in web_triggers}
            paths = {t.path for t in web_triggers if t.path}
            if len(paths) == 1:
                path = next(iter(paths))
            elif len(paths) > 1:
                raise ServeyError(
                    f"multi_paths_not_supported:{action.name}:{','.join(paths)}"
                )
            else:
                path = self.path_pattern.format(
                    action_name=action.name.replace("_", "-")
                )
            result_type = inspect.signature(action.fn).return_annotation
            if result_type == inspect.Signature.empty:
                raise ServeyError(f"missing_return_type:{action.fn}")
            endpoint = ActionEndpoint(
                action=action,
                path=path,
                methods=tuple(methods),
                params_marshaller=get_marshaller_for_params(
                    action.fn, skip_args, self.marshaller_context
                ),
                params_schema=get_schema_for_params(
                    action.fn, skip_args, self.schema_context
                ),
                result_marshaller=self.marshaller_context.get_marshaller(result_type),
                result_schema=self.schema_context.schema_from_type(result_type)
                if self.validate_output
                else None,
            )
            return endpoint



import importlib
import logging
import pkgutil
from dataclasses import dataclass, is_dataclass, field
from typing import Iterator

from servey.action.action import Action, get_action
from servey.finder.action_finder_abc import ActionFinderABC
from servey.util import get_servey_main

LOGGER = logging.getLogger(__name__)


@dataclass
class ModuleActionFinder(ActionFinderABC):
    """
    Default implementation of action_ finder which searches for actions in a particular module
    """

    root_module_name: str = field(
        default_factory=lambda: f"{get_servey_main()}.actions"
    )

    def find_actions(self) -> Iterator[Action]:
        try:
            module = importlib.import_module(self.root_module_name)
            # noinspection PyTypeChecker
            yield from _find_actions_in_module(module)
        except ModuleNotFoundError:
            LOGGER.warning("error_finding_actions")


def _find_actions_in_module(module) -> Iterator[Action]:
    for name, value in module.__dict__.items():
        action = get_action(value)
        if action:
            yield action
        if type(value) == type and is_dataclass(value):
            for param_name, param_value in value.__dict__.items():
                action = get_action(param_value)
                if action:
                    yield action
    if not hasattr(module, "__path__"):
        return  # Module was not a package...
    paths = []
    paths.extend(module.__path__)
    module_infos = list(pkgutil.walk_packages(paths))
    for module_info in module_infos:
        sub_module_name = module.__name__ + "." + module_info.name
        sub_module = importlib.import_module(sub_module_name)
        # noinspection PyTypeChecker
        yield from _find_actions_in_module(sub_module)



from dataclasses import dataclass
from typing import List, Optional, Set

from servey.action.action import Action
from servey.servey_starlette.action_endpoint.action_endpoint_abc import (
    ActionEndpointABC,
)
from servey.servey_starlette.action_endpoint.caching_action_endpoint import (
    CachingActionEndpoint,
)
from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory_abc import (
    ActionEndpointFactoryABC,
)


@dataclass
class CachingActionEndpointFactory(ActionEndpointFactoryABC):
    priority: int = 150
    skip: bool = False

    def create(
        self,
        action: Action,
        skip_args: Set[str],
        factories: List[ActionEndpointFactoryABC],
    ) -> Optional[ActionEndpointABC]:
        if self.skip or not action.cache_control:
            return
        action_endpoint = self._get_wrapped_endpoint(action, skip_args, factories)
        if action_endpoint:
            return CachingActionEndpoint(action_endpoint)

    def _get_wrapped_endpoint(
        self,
        action: Action,
        skip_args: Set[str],
        factories: List[ActionEndpointFactoryABC],
    ) -> Optional[ActionEndpointABC]:
        self.skip = True
        try:
            for factory in factories:
                action_endpoint = factory.create(action, skip_args, factories)
                if action_endpoint:
                    return action_endpoint
        finally:
            self.skip = False



import os
from logging import getLogger
from typing import Optional

from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import AuthorizerFactoryABC

LOGGER = getLogger(__name__)


class KmsAuthorizerFactory(AuthorizerFactoryABC):
    """
    Factory for authorizers that requires that boto3 is available and that
    an appropriate key exists in the kms
    """

    priority = 80

    def create_authorizer(self) -> Optional[AuthorizerABC]:
        # First we need a key id - this is stored as an environment variable
        kms_key_id = os.environ.get("KMS_KEY_ID")
        if kms_key_id is None:
            LOGGER.debug("KMS_SECRET_KEY NOT DEFINED - SKIPPING...")
            return
        from servey.servey_aws.authorizer.kms_authorizer import KmsAuthorizer

        authorizer = KmsAuthorizer("alias/" + kms_key_id)
        return authorizer



import asyncio
import json
from datetime import datetime
from typing import List, Optional
from unittest import TestCase

from starlette.datastructures import Headers
from starlette.exceptions import HTTPException
from starlette.requests import Request, empty_receive

from servey.action.action import action, get_action
from servey.action.example import Example

# noinspection PyProtectedMember
from servey.servey_starlette.action_endpoint.action_endpoint import (
    _get_nullable_schema,
    _get_valid_openapi_param_schema,
)
from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory import (
    ActionEndpointFactory,
)
from servey.trigger.web_trigger import WEB_GET, WEB_POST, WebTrigger, WebTriggerMethod
from tests.servey_strawberry.test_schema_factory import Node


class TestActionEndpoint(TestCase):
    def test_valid_input_get(self):
        @action(triggers=(WEB_GET,))
        def echo_get(val: str) -> str:
            return val

        action_endpoint = ActionEndpointFactory().create(
            get_action(echo_get), set(), []
        )
        # noinspection PyTypeChecker
        request = build_request(query_string=b"val=bar")
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(action_endpoint.execute(request))
        self.assertEqual(200, response.status_code)
        self.assertEqual("bar", json.loads(response.body))

    def test_valid_input_post(self):
        @action(triggers=(WEB_POST,))
        async def echo_get(val: str) -> str:
            return val

        action_endpoint = ActionEndpointFactory().create(
            get_action(echo_get), set(), []
        )
        request = build_request(method="POST", body=json.dumps(dict(val="bar")))
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(action_endpoint.execute(request))
        self.assertEqual(200, response.status_code)
        self.assertEqual("bar", json.loads(response.body))

    def test_invalid_input(self):
        # noinspection PyUnusedLocal
        @action(triggers=(WEB_GET,))
        def echo_get(val: int) -> int:
            """This is never invoked"""

        action_endpoint = ActionEndpointFactory().create(
            get_action(echo_get), set(), []
        )
        request = build_request(query_string="val=bar")
        loop = asyncio.get_event_loop()
        with self.assertRaises(HTTPException):
            loop.run_until_complete(action_endpoint.execute(request))

    def test_invalid_input_post(self):
        # noinspection PyUnusedLocal
        @action(triggers=(WEB_POST,))
        def echo_get(val: int) -> int:
            """This is never invoked"""

        action_endpoint = ActionEndpointFactory().create(
            get_action(echo_get), set(), []
        )
        request = build_request(method="POST", body="""{"val":"bar"}""")
        loop = asyncio.get_event_loop()
        with self.assertRaises(HTTPException):
            loop.run_until_complete(action_endpoint.execute(request))

    def test_invalid_output(self):
        @action(triggers=(WEB_GET,))
        def echo_get() -> int:
            # noinspection PyTypeChecker
            return "foobar"

        action_endpoint = ActionEndpointFactory().create(
            get_action(echo_get), set(), []
        )
        request = build_request()
        loop = asyncio.get_event_loop()
        with self.assertRaises(HTTPException):
            loop.run_until_complete(action_endpoint.execute(request))
        action_endpoint.result_schema = None
        loop.run_until_complete(action_endpoint.execute(request))

    def test_get_route(self):
        @action(triggers=(WEB_GET,))
        def dummy() -> datetime:
            """dummy route"""

        action_ = get_action(dummy)
        action_endpoint = ActionEndpointFactory().create(action_, set(), [])
        route = action_endpoint.get_route()
        self.assertEqual("dummy", route.name)
        self.assertEqual("/actions/dummy", route.path)
        self.assertEqual({"GET", "HEAD"}, set(route.methods))

    def test_open_api_get(self):
        # noinspection PyUnusedLocal
        @action(
            triggers=(WEB_GET,),
            examples=(
                Example(
                    "standard",
                    params=dict(id="id", value=1.2, count=3, flag=True),
                    result="2022-12-01 00:00:00+00:00",
                    description="Standard invocation of dummy",
                ),
            ),
        )
        def dummy(id: str, value: float, count: int, flag: bool = True) -> datetime:
            """dummy route"""

        action_ = get_action(dummy)
        action_endpoint = ActionEndpointFactory().create(action_, set(), [])
        schema = dict(paths={}, components={})
        action_endpoint.to_openapi_schema(schema)
        expected_schema = {
            "components": {},
            "paths": {
                "/actions/dummy": {
                    "get": {
                        "operationId": "dummy",
                        "parameters": [
                            {
                                "examples": {
                                    "standard": {
                                        "summary": "Standard "
                                        "invocation "
                                        "of "
                                        "dummy",
                                        "value": "id",
                                    }
                                },
                                "in": "query",
                                "name": "id",
                                "required": True,
                                "schema": {"type": "string"},
                            },
                            {
                                "examples": {
                                    "standard": {
                                        "summary": "Standard "
                                        "invocation "
                                        "of "
                                        "dummy",
                                        "value": 1.2,
                                    }
                                },
                                "in": "query",
                                "name": "value",
                                "required": True,
                                "schema": {"type": "number"},
                            },
                            {
                                "examples": {
                                    "standard": {
                                        "summary": "Standard "
                                        "invocation "
                                        "of "
                                        "dummy",
                                        "value": 3,
                                    }
                                },
                                "in": "query",
                                "name": "count",
                                "required": True,
                                "schema": {"type": "integer"},
                            },
                            {
                                "examples": {
                                    "standard": {
                                        "summary": "Standard "
                                        "invocation "
                                        "of "
                                        "dummy",
                                        "value": True,
                                    }
                                },
                                "in": "query",
                                "name": "flag",
                                "required": False,
                                "schema": {"type": "boolean"},
                            },
                        ],
                        "responses": {
                            "200": {
                                "content": {
                                    "application/json": {
                                        "examples": {
                                            "standard": {
                                                "summary": "Standard "
                                                "invocation "
                                                "of "
                                                "dummy",
                                                "value": "2022-12-01 " "00:00:00+00:00",
                                            }
                                        },
                                        "schema": {
                                            "format": "date-time",
                                            "type": "string",
                                        },
                                    }
                                },
                                "description": "Successful " "Response",
                            },
                            "422": {"description": "Validation " "Error"},
                        },
                        "summary": "dummy route",
                    }
                }
            },
        }
        self.assertEqual(expected_schema, schema)

    def test_open_api_post(self):
        # noinspection PyUnusedLocal
        @action(
            triggers=(WEB_POST,),
            examples=(
                Example(
                    "standard",
                    params=dict(node=dict(name="Foo")),
                    result=dict(name="Foo", child_nodes=[]),
                    description="Standard invocation of dummy",
                ),
            ),
        )
        def dummy(node: Node) -> Node:
            """dummy route"""

        action_ = get_action(dummy)
        action_endpoint = ActionEndpointFactory().create(action_, set(), [])
        schema = dict(paths={}, components={})
        action_endpoint.to_openapi_schema(schema)
        expected_schema = {
            "components": {
                "Node": {
                    "additionalProperties": False,
                    "description": "Node(name: str, child_nodes: "
                    "List[ForwardRef('tests.servey_strawberry.test_schema_factory.Node')] "
                    "= <factory>)",
                    "name": "Node",
                    "properties": {
                        "child_nodes": {
                            "items": {"$ref": "#/components/Node"},
                            "type": "array",
                        },
                        "name": {"type": "string"},
                    },
                    "required": ["name"],
                    "type": "object",
                }
            },
            "paths": {
                "/actions/dummy": {
                    "post": {
                        "operationId": "dummy",
                        "requestBody": {
                            "content": {
                                "application/json": {
                                    "examples": {
                                        "standard": {
                                            "summary": "Standard "
                                            "invocation "
                                            "of "
                                            "dummy",
                                            "value": {"node": {"name": "Foo"}},
                                        }
                                    },
                                    "schema": {
                                        "additionalProperties": True,
                                        "properties": {
                                            "node": {"$ref": "#/components/Node"}
                                        },
                                        "required": ["node"],
                                        "type": "object",
                                    },
                                }
                            },
                            "required": True,
                        },
                        "responses": {
                            "200": {
                                "content": {
                                    "application/json": {
                                        "examples": {
                                            "standard": {
                                                "summary": "Standard "
                                                "invocation "
                                                "of "
                                                "dummy",
                                                "value": {
                                                    "child_nodes": [],
                                                    "name": "Foo",
                                                },
                                            }
                                        },
                                        "schema": {"$ref": "#/components/Node"},
                                    }
                                },
                                "description": "Successful " "Response",
                            },
                            "422": {"description": "Validation " "Error"},
                        },
                        "summary": "dummy route",
                    }
                }
            },
        }
        self.assertEqual(expected_schema, schema)

    def test_nested_param_get(self):
        # noinspection PyUnusedLocal
        @action(
            triggers=(WEB_GET,),
        )
        def dummy(node: Node) -> str:
            """dummy route"""

        action_ = get_action(dummy)
        action_endpoint = ActionEndpointFactory().create(action_, set(), [])
        schema = dict(paths={}, components={})
        action_endpoint.to_openapi_schema(schema)
        expected_schema = {
            "paths": {
                "/actions/dummy": {
                    "get": {
                        "responses": {
                            "422": {"description": "Validation Error"},
                            "200": {
                                "description": "Successful Response",
                                "content": {
                                    "application/json": {"schema": {"type": "string"}}
                                },
                            },
                        },
                        "operationId": "dummy",
                        "summary": "dummy route",
                        "parameters": [
                            {
                                "required": True,
                                "schema": {"type": "string"},
                                "name": "node.name",
                                "in": "query",
                            }
                        ],
                    }
                }
            },
            "components": {},
        }
        self.assertEqual(expected_schema, schema)

    def test_example_missing_parameters(self):
        # noinspection PyUnusedLocal
        @action(
            triggers=(WEB_GET,),
            examples=(
                Example("has_b", {"a": 1, "b": 2}, result=3),
                Example("missing_b", {"a": 1}, result=2),
            ),
        )
        def add(a: int, b: int = None) -> int:
            """Dummy"""

        endpoint = ActionEndpointFactory().create(get_action(add), set(), [])
        schema = {"paths": {}, "components": {}}
        endpoint.to_openapi_schema(schema)
        expected_schema = {
            "paths": {
                "/actions/add": {
                    "get": {
                        "responses": {
                            "422": {"description": "Validation Error"},
                            "200": {
                                "description": "Successful Response",
                                "content": {
                                    "application/json": {
                                        "schema": {"type": "integer"},
                                        "examples": {
                                            "has_b": {"value": 3},
                                            "missing_b": {"value": 2},
                                        },
                                    }
                                },
                            },
                        },
                        "operationId": "add",
                        "summary": "Dummy",
                        "parameters": [
                            {
                                "required": True,
                                "schema": {"type": "integer"},
                                "name": "a",
                                "in": "query",
                                "examples": {
                                    "has_b": {"value": 1},
                                    "missing_b": {"value": 1},
                                },
                            },
                            {
                                "required": False,
                                "schema": {"type": "integer"},
                                "name": "b",
                                "in": "query",
                                "examples": {"has_b": {"value": 2}},
                            },
                        ],
                    }
                }
            },
            "components": {},
        }
        self.assertEqual(expected_schema, schema)

    def test_example_array_parameters(self):
        # noinspection PyUnusedLocal
        @action(
            triggers=(WEB_GET,),
            examples=(Example("standard", {"values": [1, 2, 3]}, result=6),),
        )
        def add(values: List[int]) -> int:
            """Dummy"""

        endpoint = ActionEndpointFactory().create(get_action(add), set(), [])
        schema = {"paths": {}, "components": {}}
        endpoint.to_openapi_schema(schema)
        expected_schema = {
            "paths": {
                "/actions/add": {
                    "get": {
                        "responses": {
                            "422": {"description": "Validation Error"},
                            "200": {
                                "description": "Successful Response",
                                "content": {
                                    "application/json": {
                                        "schema": {"type": "integer"},
                                        "examples": {
                                            "standard": {"value": 6},
                                        },
                                    }
                                },
                            },
                        },
                        "operationId": "add",
                        "summary": "Dummy",
                        "parameters": [
                            {
                                "required": True,
                                "schema": {
                                    "type": "array",
                                    "items": {"type": "integer"},
                                },
                                "name": "values",
                                "in": "query",
                                "examples": {"standard": {"value": [1, 2, 3]}},
                            }
                        ],
                    }
                }
            },
            "components": {},
        }
        self.assertEqual(expected_schema, schema)

    def test_example_array_object_parameters(self):
        # noinspection PyUnusedLocal
        @action(triggers=(WEB_GET,))
        def add(values: Optional[Node]) -> int:
            """Dummy"""

        endpoint = ActionEndpointFactory().create(get_action(add), set(), [])
        schema = {"paths": {}, "components": {}}
        endpoint.to_openapi_schema(schema)
        expected_schema = {
            "components": {},
            "paths": {
                "/actions/add": {
                    "get": {
                        "operationId": "add",
                        "parameters": [
                            {
                                "in": "query",
                                "name": "values.name",
                                "required": True,
                                "schema": {"type": "string"},
                            }
                        ],
                        "responses": {
                            "200": {
                                "content": {
                                    "application/json": {"schema": {"type": "integer"}}
                                },
                                "description": "Successful " "Response",
                            },
                            "422": {"description": "Validation " "Error"},
                        },
                        "summary": "Dummy",
                    }
                }
            },
        }
        self.assertEqual(expected_schema, schema)

    def test_nullable_schema(self):
        self.assertIsNone(_get_nullable_schema({"anyOf": "invalid"}))
        self.assertIsNone(
            _get_nullable_schema(
                {
                    "anyOf": [
                        {"type": "string"},
                        {"type": "int"},
                    ]
                }
            )
        )

    def test_get_valid_openapi_param_schema(self):
        self.assertIsNone(_get_valid_openapi_param_schema({}))
        self.assertIsNone(_get_valid_openapi_param_schema({"type": "array"}))

    def test_get_with_path_params(self):
        # noinspection PyUnusedLocal
        @action(triggers=(WebTrigger(WebTriggerMethod.GET, "/find/{key}"),))
        def find(key: str) -> int:
            """ Dummy """
            return 17

        endpoint = ActionEndpointFactory().create(get_action(find), set(), [])
        schema = {"paths": {}, "components": {}}
        endpoint.to_openapi_schema(schema)
        expected_schema = {
            "components": {},
            "paths": {
                "/find/{key}": {
                    "get": {
                        "operationId": "find",
                        "parameters": [
                            {
                                "in": "path",
                                "name": "key",
                                "required": True,
                                "schema": {"type": "string"},
                            }
                        ],
                        "responses": {
                            "200": {
                                "content": {
                                    "application/json": {"schema": {"type": "integer"}}
                                },
                                "description": "Successful " "Response",
                            },
                            "422": {"description": "Validation " "Error"},
                        },
                        "summary": "Dummy",
                    }
                }
            },
        }
        self.assertEqual(expected_schema, schema)
        request = build_request(path_params={'key': 'bar'})
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(endpoint.execute(request))
        self.assertEqual(200, response.status_code)
        self.assertEqual(17, json.loads(response.body))

    def test_post_with_path_params(self):
        # noinspection PyUnusedLocal
        @action(triggers=(WebTrigger(WebTriggerMethod.POST, "/find/{key}"),))
        def find(key: str, value: str) -> int:
            """ Dummy """
            return len(key) + len(value)

        endpoint = ActionEndpointFactory().create(get_action(find), set(), [])
        schema = {"paths": {}, "components": {}}
        endpoint.to_openapi_schema(schema)
        expected_schema = {
            "components": {},
            "paths": {
                "/find/{key}": {
                    "post": {
                        "operationId": "find",
                        "parameters": [
                            {
                                "in": "path",
                                "name": "key",
                                "required": True,
                                "schema": {"type": "string"},
                            }
                        ],
                        'requestBody': {
                            'content': {
                                'application/json': {
                                    'schema': {
                                        'additionalProperties': True,
                                        'properties': {
                                            'value': {'type': 'string'}
                                        },
                                        'required': ['value'],
                                        'type': 'object'
                                    }
                                }
                            },
                            'required': True
                        },
                        "responses": {
                            "200": {
                                "content": {
                                    "application/json": {"schema": {"type": "integer"}}
                                },
                                "description": "Successful " "Response",
                            },
                            "422": {"description": "Validation " "Error"},
                        },
                        "summary": "Dummy",
                    }
                }
            },
        }
        self.assertEqual(expected_schema, schema)
        request = build_request(method='POST', path_params={'key': 'bar'}, body=json.dumps({"value": "ping"}))
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(endpoint.execute(request))
        self.assertEqual(200, response.status_code)
        self.assertEqual(7, json.loads(response.body))


def build_request(
    method: str = "GET",
    server: str = "www.example.com",
    path: str = "/",
    query_string: str = None,
    headers: dict = None,
    body: str = None,
    path_params: dict = None,
) -> Request:
    if headers is None:
        headers = {}
    receive = empty_receive
    if body:

        async def receive_body():
            return {"type": "http.request", "body": body.encode()}

        receive = receive_body

    request = Request(
        {
            "type": "http",
            "path": path,
            "headers": Headers(headers).raw,
            "http_version": "1.1",
            "method": method,
            "scheme": "https",
            "client": ("127.0.0.1", 8080),
            "server": (server, 443),
            "query_string": query_string,
            "path_params": path_params,
        },
        receive=receive,
    )
    return request



import dataclasses
import importlib
import inspect
import json
import logging
import os
from marshy.types import ExternalItemType, ExternalType

from servey.action.action import get_action
from servey.errors import ServeyError
from servey.servey_aws.event_handler.event_handler_abc import get_event_handlers


def invoke(event: ExternalItemType, context) -> ExternalType:
    _LOGGER.info(json.dumps(dict(lambda_event=event)))
    for handler in _EVENT_HANDLERS:
        if handler.is_usable(event, context):
            return handler.handle(event, context)
    raise ServeyError("no_handler")


def find_action():
    action_module = importlib.import_module(os.environ["SERVEY_ACTION_MODULE"])
    action_function_name = os.environ["SERVEY_ACTION_FUNCTION_NAME"]
    if "." in action_function_name:
        action_class_name, action_function_name = action_function_name.split(".")
        action_class = getattr(action_module, action_class_name)
        action_function = getattr(action_class, action_function_name)
        action_ = get_action(action_function)
        fn = action_.fn

        def wrapper(*args, **kwargs):
            result = fn(*args, **kwargs)
            return result

        sig = inspect.signature(fn)
        sig = sig.replace(
            parameters=[
                p.replace(annotation=action_class) if p.name == "self" else p
                for p in sig.parameters.values()
            ]
        )
        wrapper.__signature__ = sig
        action_ = dataclasses.replace(action_, fn=wrapper)
        return action_
    else:
        action_function = getattr(action_module, action_function_name)
        action_ = get_action(action_function)
        return action_


_ACTION = find_action()
_EVENT_HANDLERS = get_event_handlers(_ACTION)
_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)
logging.basicConfig(level=logging.INFO)



from datetime import datetime, timezone
from unittest import TestCase

from servey.util.to_second_datetime_marshaller import ToSecondDatetimeMarshaller


class TestToSecondDatetimeMarshaller(TestCase):
    def test_load_and_dump(self):
        time = datetime(2022, 12, 1, 11, 43, 26, tzinfo=timezone.utc)
        marshaller = ToSecondDatetimeMarshaller()
        dumped = marshaller.dump(time)
        self.assertEqual("2022-12-01T11:43:26+00:00", dumped)
        loaded = marshaller.load(dumped)
        self.assertEqual(time, loaded)

    def test_dump_microseconds(self):
        time = datetime(2020, 2, 3, 4, 5, 6, microsecond=503, tzinfo=timezone.utc)
        marshaller = ToSecondDatetimeMarshaller()
        dumped = marshaller.dump(time)
        self.assertEqual("2020-02-03T04:05:06+00:00", dumped)



from dataclasses import dataclass
from typing import Type

from marshy.types import ExternalItemType

from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.authorization import AuthorizationError
from servey.servey_aws.event_handler.event_handler import (
    EventHandler,
    EventHandlerFactory,
)
from servey.servey_aws.event_handler.event_handler_abc import EventHandlerABC


class AppsyncEventHandler(EventHandler):
    def is_usable(self, event: ExternalItemType, context) -> bool:
        return "arguments" in event

    def parse_kwargs(self, event: ExternalItemType):
        arguments = dict(**event["arguments"])
        source = event.get("source")
        if source is not None:
            arguments["self"] = source
        if self.param_schema:
            self.param_schema.validate(arguments)
        kwargs = self.param_marshaller.load(arguments)
        if self.auth_kwarg_name or self.action.access_control != ALLOW_ALL:
            # noinspection PyTypeChecker
            headers: ExternalItemType = (event.get("request") or {}).get(
                "headers"
            ) or {}
            auth_header = headers.get("authorization")
            authorization = None
            if auth_header:
                if auth_header.lower().startswith("bearer "):
                    auth_header = auth_header[7:]
                authorization = self.authorizer.authorize(auth_header)
            if not self.action.access_control.is_executable(authorization):
                raise AuthorizationError("unauthorized")
            if self.auth_kwarg_name:
                kwargs[self.auth_kwarg_name] = authorization
        return kwargs


@dataclass
class AppsyncEventHandlerFactory(EventHandlerFactory):
    priority: int = 100
    event_handler_type: Type[EventHandlerABC] = AppsyncEventHandler



from marshy.types import ExternalItemType

from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    ensure_ref_in_file,
    create_yml_file,
)


class KmsKeyConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    kms_resource_yml_file: str = "serverless_servey/kms_resource.yml"
    kms_role_statement_yml_file: str = "serverless_servey/kms_role_statement.yml"

    def configure(self, main_serverless_yml_file: str):
        ensure_ref_in_file(
            main_serverless_yml_file,
            ["resources"],
            self.kms_resource_yml_file,
        )
        ensure_ref_in_file(
            main_serverless_yml_file,
            ["provider", "iamRoleStatements"],
            self.kms_role_statement_yml_file,
            "iamRoleStatements.0",
        )
        kms_resource_yml = self.build_kms_resource_yml()
        create_yml_file(
            self.kms_resource_yml_file, kms_resource_yml, _mutate_kms_resource_str
        )
        kms_role_statement_yml = self.build_kms_role_statement_yml()
        create_yml_file(self.kms_role_statement_yml_file, kms_role_statement_yml)

    @staticmethod
    def build_kms_resource_yml() -> ExternalItemType:
        kms_resource = {
            "Resources": {
                "serveyKmsKey": {
                    "Type": "AWS::KMS::Key",
                    "Properties": {
                        "Description": "Key used for servey authentication and authorization",
                        "Enabled": True,
                        "KeySpec": "RSA_4096",
                        "KeyUsage": "SIGN_VERIFY",
                        "KeyPolicy": {
                            "Version": "2012-10-17",
                            "Id": "serveyKmsKey",
                            "Statement": {
                                "Sid": "Allow administration of the key",
                                "Effect": "Allow",
                                "Principal": {
                                    "AWS": "AWS_ACCOUNT_ID",
                                },
                                "Action": "kms:*",
                                "Resource": "*",
                            },
                        },
                    },
                },
                "serveyKmsKeyAlias": {
                    "Type": "AWS::KMS::Alias",
                    "Properties": {
                        "AliasName": "alias/serveyKmsKey",
                        "TargetKeyId": "SERVEY_KMS_KEY",
                    },
                },
            },
        }
        return kms_resource

    @staticmethod
    def build_kms_role_statement_yml() -> ExternalItemType:
        kms_policy = {
            "iamRoleStatements": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "kms:Sign",
                        "kms:Verify",
                        "kms:GetPublicKey",
                    ],
                    "Resource": {
                        "Fn::GetAtt": [
                            "serveyKmsKey",
                            "Arn",
                        ]
                    },
                }
            ]
        }
        return kms_policy


def _mutate_kms_resource_str(kms_resource: str) -> str:
    kms_resource = kms_resource.replace(
        "AWS_ACCOUNT_ID", "!Sub arn:aws:iam::${aws:accountId}:root"
    )
    kms_resource = kms_resource.replace("SERVEY_KMS_KEY", "!Ref serveyKmsKey")
    return kms_resource



import asyncio
import os
from unittest import TestCase
from unittest.mock import patch

from servey.action.action import action, get_action
from servey.servey_thread.asyncio_subscription_service import (
    AsyncioSubscriptionServiceFactory,
)
from servey.subscription.subscription import subscription


class TestAsyncioSubscriptionService(TestCase):
    def test_subscriptions(self):
        total = 0

        @action
        def accumulator(value: int):
            nonlocal total
            total += value

        subscription_ = subscription(
            event_type=int,
            name="accumulators",
            action_subscribers=(get_action(accumulator),),
        )
        factory = AsyncioSubscriptionServiceFactory()
        service = factory.create([subscription_])
        service.publish(subscription_, 3)
        service.publish(subscription_, 5)
        service.publish(subscription_, 7)
        loop = asyncio.get_event_loop()

        async def dummy():
            pass

        loop.run_until_complete(dummy())
        self.assertEqual(15, total)

    def test_disabled_with_lambda(self):
        with patch.dict(os.environ, {"AWS_REGION": "bezos-moon-base"}):
            factory = AsyncioSubscriptionServiceFactory()
            subscription_ = subscription(event_type=int, name="accumulators")
            self.assertIsNone(factory.create([subscription_]))



from datetime import datetime
from unittest import TestCase

from servey.cache_control.cache_header import CacheHeader


class TestCacheHeader(TestCase):
    def test_cache_control_str_no_expire(self):
        cache_header = CacheHeader("etag")
        self.assertEqual("no-storage", cache_header.get_cache_control_str())

    def test_combine_with(self):
        a = CacheHeader(
            "a",
            datetime.fromisoformat("2020-01-01"),
            datetime.fromisoformat("2020-06-01"),
        )
        b = CacheHeader(
            "b",
            datetime.fromisoformat("2020-02-01"),
            datetime.fromisoformat("2020-05-01"),
        )
        c = a.combine_with([b])
        d = b.combine_with([a])
        self.assertIsNotNone(c.etag)
        self.assertIsNotNone(d.etag)
        self.assertNotEqual(d.etag, c.etag)
        e = CacheHeader(
            c.etag,
            datetime.fromisoformat("2020-02-01"),
            datetime.fromisoformat("2020-05-01"),
        )
        self.assertEqual(e, c)
        f = CacheHeader(
            d.etag,
            datetime.fromisoformat("2020-02-01"),
            datetime.fromisoformat("2020-05-01"),
        )
        self.assertEqual(f, d)



from __future__ import annotations
import json
import logging
from dataclasses import dataclass
from string import Formatter
from typing import Tuple, Any, Optional, Dict, List, Iterator, Awaitable

from json_urley import query_str_to_json_obj
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType
from schemey import Schema
from schemey.util import filter_none
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.routing import Route

from servey.action.action import Action
from servey.action.example import Example
from servey.action.util import move_ref_items_to_components
from servey.servey_starlette.action_endpoint.action_endpoint_abc import (
    ActionEndpointABC,
)
from servey.trigger.web_trigger import WebTriggerMethod, BODY_METHODS

LOGGER = logging.getLogger(__name__)


@dataclass
class ActionEndpoint(ActionEndpointABC):
    """
    Wrapper for a function for use within starlette, with everything needed to bind it to a route
    """

    action: Action
    path: str
    methods: Tuple[WebTriggerMethod, ...]
    params_marshaller: MarshallerABC
    params_schema: Optional[Schema]
    result_marshaller: MarshallerABC
    result_schema: Optional[Schema] = None

    def __post_init__(self):
        self.field_names = {
            fname for _, fname, _, _ in Formatter().parse(self.path) if fname
        }

    def get_action(self) -> Action:
        return self.action

    def get_route(self) -> Route:
        return Route(
            self.path,
            name=self.action.name,
            endpoint=self.execute,
            methods=[m.value for m in self.methods],
        )

    async def execute_with_context(
        self, request: Request, context: Dict[str, Any]
    ) -> Response:
        kwargs = await self.parse_request(request)
        kwargs.update(context)
        result = self.action.fn(**kwargs)
        if isinstance(result, Awaitable):
            result = await result
        # Lazy action resolution would be done here!
        response = self.render_response(result)
        return response

    async def parse_request(self, request: Request):
        method = WebTriggerMethod(request.method.lower())
        if method in BODY_METHODS:
            body = await request.body()
            params: ExternalItemType = json.loads(body) if body else {}
            if request.path_params:
                params.update(request.path_params)
            error = next(self.params_schema.iter_errors(params), None)
            if error:
                raise HTTPException(422, str(error))
            kwargs = self.params_marshaller.load(params)
        else:
            try:
                query_str = request.scope["query_string"] or b""
                if isinstance(query_str, bytes):
                    query_str = query_str.decode("latin-1")
                json_obj = query_str_to_json_obj(query_str)
                if request.path_params:
                    json_obj.update(request.path_params)
                self.params_schema.validate(json_obj)
                kwargs = self.params_marshaller.load(json_obj)
            except Exception:
                LOGGER.exception("invalid_input")
                raise HTTPException(422, "invalid_input")
        return kwargs

    def render_response(self, result: Any):
        result_content = self.result_marshaller.dump(result)
        if self.result_schema:
            error = next(self.result_schema.iter_errors(result_content), None)
            if error:
                raise HTTPException(500, str(error))
        return JSONResponse(result_content)

    def to_openapi_schema(self, schema: ExternalItemType):
        paths: ExternalItemType = schema["paths"]
        components: ExternalItemType = schema["components"]
        path = paths.get(self.path)
        if not path:
            path = paths[self.path] = {}
        for method in self.methods:
            responses = {}
            path_method = path[method.value] = {"responses": responses}
            path_method["operationId"] = self.action.name
            if self.action.description:
                path_method["summary"] = self.action.description
            # Tags?
            self.request_to_openapi_schema(method, path_method, components)
            self.response_to_openapi_schema(responses, components)

    def request_to_openapi_schema(
        self,
        method: WebTriggerMethod,
        path_method: ExternalItemType,
        components: ExternalItemType,
    ):
        if method in BODY_METHODS:
            self.body_request_to_openapi_schema(path_method, components)
        else:
            self.query_string_request_to_openapi_schema(path_method)

    def body_request_to_openapi_schema(
        self, path_method: ExternalItemType, components: ExternalItemType
    ):
        schema = move_ref_items_to_components(
            self.params_schema.schema, self.params_schema.schema, components
        )

        if self.field_names:
            parameters = []
            properties = schema.get("properties")
            required_properties = schema.get("required") or []
            for field_name in self.field_names:
                parameters.append(
                    {
                        "required": field_name in required_properties,
                        "schema": properties.pop(field_name),
                        "name": field_name,
                        "in": "path",
                    }
                )
                if field_name in required_properties:
                    required_properties.remove(field_name)
            path_method["parameters"] = parameters

        # move params from schema to params if path params
        content = {"schema": schema}
        path_method["requestBody"] = {
            "content": {"application/json": content},
            "required": True,
        }
        if self.action.examples:
            content["examples"] = {
                e.name: filter_none(dict(summary=e.description, value=e.params))
                for e in self.action.examples
                if e.include_in_schema
            }
        responses: ExternalItemType = path_method["responses"]
        responses["422"] = {"description": "Validation Error"}

    def query_string_request_to_openapi_schema(self, path_method: ExternalItemType):
        # Schema to openapi
        params = list(
            _object_schema_to_json_urley_parameters(
                object_schema=self.params_schema.schema,
                examples=self.action.examples or tuple(),
                path=[],
            )
        )
        if self.field_names:
            for param in params:
                if param["name"] in self.field_names:
                    param["in"] = "path"
        path_method["parameters"] = params
        responses: ExternalItemType = path_method["responses"]
        responses["422"] = {"description": "Validation Error"}

    def response_to_openapi_schema(
        self, responses: ExternalItemType, components: ExternalItemType
    ):
        schema = self.result_schema.schema
        schema = move_ref_items_to_components(schema, schema, components)
        content = {"schema": schema}
        if self.action.examples:
            content["examples"] = {
                e.name: filter_none(dict(summary=e.description, value=e.result))
                for e in self.action.examples
                if e.include_in_schema
            }
        responses["200"] = {
            "description": "Successful Response",
            "content": {"application/json": content},
        }


def _object_schema_to_json_urley_parameters(
    object_schema: ExternalItemType, examples: Tuple[Example], path: List[str]
) -> Iterator[ExternalItemType]:
    """
    This is not full support for json urley - parameters with references / urls are ignored and just do not appear
    in the schema.
    """
    properties: ExternalItemType = object_schema.get("properties")
    required_properties = set(object_schema.get("required") or [])
    for key, property_schema in properties.items():
        valid_schema = _get_valid_openapi_param_schema(property_schema)
        if not valid_schema:
            continue
        path.append(key.replace("~", "~~").replace(".", "~."))
        valid_schema = _get_nullable_schema(valid_schema) or valid_schema
        if valid_schema.get("type") == "object":
            yield from _object_schema_to_json_urley_parameters(
                valid_schema, examples, path
            )
        else:
            example_schemas = {}
            for example in examples:
                example_value = example.params
                for p in path:
                    example_value = example_value.get(p, None)
                    if example_value is None:
                        break
                if example_value is not None:
                    example_schema = dict(value=example_value)
                    if example.description:
                        example_schema["summary"] = example.description
                    example_schemas[example.name] = example_schema

            result = {
                "required": key in required_properties,
                "schema": property_schema,
                "name": ".".join(path),
                "in": "query",
            }
            if example_schemas:
                result["examples"] = example_schemas
            yield result
        path.pop()


def _get_valid_openapi_param_schema(schema: ExternalItemType):
    nullable_schema = _get_nullable_schema(schema)
    if nullable_schema:
        sub_schema = _get_valid_openapi_param_schema(nullable_schema)
        return nullable_schema if sub_schema else None
    type_ = schema.get("type")
    if not type_:
        return
    type_ = type_.lower()
    if type_ == "array":
        sub_schema: ExternalItemType = schema.get("items")
        if not sub_schema:
            return
        sub_schema = _get_nullable_schema(sub_schema) or sub_schema
        type_ = sub_schema.get("type")
        if type_ in ("boolean", "number", "integer", "string", "null"):
            # We currently don't support nested arrays of objects / arrays in an openapi schema
            return schema
    elif type_ in ("boolean", "number", "integer", "object", "string"):
        return schema


def _get_nullable_schema(schema: ExternalItemType):
    any_of = schema.get("anyOf")
    if any_of:
        if not isinstance(any_of, list):
            return
        non_nulls = [s for s in any_of if s.get("type") != "null"]
        if len(non_nulls) != 1:
            return
        return non_nulls[0]



import inspect
import json
from dataclasses import field, dataclass
from typing import Optional

from marshy import get_default_context, ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType
from schemey import get_default_schema_context, SchemaContext, Schema

from servey.action.action import Action
from servey.servey_aws.event_handler.event_handler_abc import (
    EventHandlerABC,
    EventHandlerFactoryABC,
)


@dataclass
class SqsEventHandler(EventHandlerABC):
    action: Action
    event_marshaller: MarshallerABC
    event_schema: Optional[Schema]
    result_marshaller: MarshallerABC
    priority: int = 50

    def is_usable(self, event: ExternalItemType, context) -> bool:
        return "Records" in event

    def handle(self, event: ExternalItemType, context) -> ExternalType:
        # noinspection PyTypeChecker
        records = [json.loads(r["body"]) for r in event["Records"]]
        if self.event_schema:
            for r in records:
                self.event_schema.validate(r)
        arg_list = [self.event_marshaller.load(r) for r in records]
        if self.action.batch_invoker:
            results = self.action.batch_invoker.fn(arg_list)
        else:
            results = [self.action.fn(a) for a in arg_list]
        if self.result_marshaller:
            results = [self.result_marshaller.dump(r) for r in results]
        return results


@dataclass
class SqsEventHandlerFactory(EventHandlerFactoryABC):
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    schema_context: SchemaContext = field(default_factory=get_default_schema_context)
    priority: int = 100

    def create(self, action: Action) -> Optional[EventHandlerABC]:
        sig = inspect.signature(action.fn)
        params = list(sig.parameters.values())
        if len(params) != 1:
            return  # Sqs requires a single event
        event_type = params[0].annotation
        event_schema = self.schema_context.schema_from_type(event_type)
        event_marshaller = self.marshaller_context.get_marshaller(event_type)
        result_marshaller = None
        if sig.return_annotation != inspect.Signature.empty:
            result_marshaller = self.marshaller_context.get_marshaller(
                sig.return_annotation
            )
        return SqsEventHandler(
            action=action,
            event_marshaller=event_marshaller,
            event_schema=event_schema,
            result_marshaller=result_marshaller,
            priority=self.priority,
        )



import os
from unittest import TestCase
from unittest.mock import patch

from servey.action.action import action, get_action
from servey.servey_aws.sqs_subscription_service import SqsSubscriptionServiceFactory
from servey.subscription.subscription import subscription


class TestSqsSubscriptionService(TestCase):
    def test_publish(self):
        # noinspection PyUnusedLocal
        @action
        def mock_consumer(value: int):
            """No action required"""

        def mock_boto3_client(name: str):
            self.assertEqual("sqs", name)
            return MockSqsClient()

        class MockSqsClient:
            @staticmethod
            def get_queue_url(QueueName: str):
                self.assertEqual("servey_main-dummy_subscription", QueueName)
                return {"QueueUrl": "https://foobar.com"}

            @staticmethod
            def send_message(QueueUrl: str, MessageBody: str):
                messages.append(dict(queue_url=QueueUrl, body=MessageBody))

        messages = []
        subscription_ = subscription(
            int, "dummy_subscription", action_subscribers=(get_action(mock_consumer),)
        )
        non_sqs_subscription = subscription(int, "non_sqs")
        with (
            patch("boto3.client", mock_boto3_client),
            patch.dict(
                os.environ,
                {
                    "AWS_REGION": "bezos-moon-base",
                    "CONNECTION_TABLE_NAME": "mock_connections",
                },
            ),
        ):
            factory = SqsSubscriptionServiceFactory()
            service = factory.create([subscription_])
            service.publish(subscription_, 13)
            service.publish(non_sqs_subscription, 14)
            self.assertEqual(
                [{"body": "13", "queue_url": "https://foobar.com"}], messages
            )



from dataclasses import field, dataclass
from typing import List

from marshy.types import ExternalItemType

from servey.finder.subscription_finder_abc import find_subscriptions
from servey.security.access_control.allow_none import ALLOW_NONE
from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    ensure_ref_in_file,
    create_yml_file,
)
from servey.subscription.subscription import Subscription
from servey.util import get_servey_main


@dataclass
class SubscriptionFunctionConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    handlers_yml_file: str = "serverless_servey/subscriptions_handlers.yml"
    resources_yml_file: str = "serverless_servey/subscriptions_resources.yml"
    role_statement_yml_file: str = "serverless_servey/subscriptions_role_statement.yml"
    subscriptions: List[Subscription] = field(
        default_factory=lambda: list(find_subscriptions())
    )

    @property
    def has_websocket_subscriptions(self):
        has_websocket_subscriptions = next(
            (True for s in self.subscriptions if s.access_control != ALLOW_NONE), False
        )
        return has_websocket_subscriptions

    @property
    def connection_table_name(self):
        connection_table_name = get_servey_main() + "_connection"
        return connection_table_name

    def configure(self, main_serverless_yml_file: str):
        if not self.subscriptions:
            # Should we remove files if missing?
            return

        if self.has_websocket_subscriptions:
            ensure_ref_in_file(
                main_serverless_yml_file,
                ["functions"],
                self.handlers_yml_file,
            )
            handlers_yml = self.build_handlers_yml()
            create_yml_file(self.handlers_yml_file, handlers_yml)

        ensure_ref_in_file(
            main_serverless_yml_file,
            ["resources"],
            self.resources_yml_file,
        )
        resources_yml = self.build_resources_yml()
        create_yml_file(self.resources_yml_file, resources_yml)

        role_statement_yml = self.build_role_statement_yml()
        create_yml_file(self.role_statement_yml_file, role_statement_yml)
        for i in range(len(role_statement_yml["iamRoleStatements"])):
            ensure_ref_in_file(
                main_serverless_yml_file,
                ["provider", "iamRoleStatements"],
                self.role_statement_yml_file,
                f"iamRoleStatements.{i}",
            )

    @staticmethod
    def build_handlers_yml() -> ExternalItemType:
        handler_definitions = {
            "subscriptions": {
                "handler": "servey.servey_aws.lambda_websocket.lambda_handler",
                "events": [
                    {"websocket": {"route": "$connect"}},
                    {"websocket": {"route": "$disconnect"}},
                    {"websocket": {"route": "$default"}},
                ],
            }
        }
        return handler_definitions

    def build_resources_yml(self) -> ExternalItemType:
        service_name = get_servey_main()
        resources = {
            s.name.title().replace("_", "")
            + "SQS": {
                "Type": "AWS::SQS::Queue",
                "Properties": {"QueueName": service_name + "-" + s.name},
            }
            for s in self.subscriptions
            if s.action_subscribers
        }
        if self.has_websocket_subscriptions:
            resources[self.connection_table_name.title().replace("_", "")] = {
                "Type": "AWS::DynamoDB::Table",
                "Properties": {
                    "TableName": self.connection_table_name,
                    "BillingMode": "PAY_PER_REQUEST",
                    "AttributeDefinitions": [
                        {"AttributeName": "connection_id", "AttributeType": "S"},
                        {
                            "AttributeName": "subscription_name",
                            "AttributeType": "S",
                        },
                    ],
                    "KeySchema": [
                        {"AttributeName": "connection_id", "KeyType": "HASH"},
                        {"AttributeName": "subscription_name", "KeyType": "RANGE"},
                    ],
                    "GlobalSecondaryIndexes": [
                        {
                            "IndexName": "gsi__subscription_name__connection_id",
                            "Projection": {"ProjectionType": "ALL"},
                            "KeySchema": [
                                {
                                    "AttributeName": "subscription_name",
                                    "KeyType": "HASH",
                                },
                                {
                                    "AttributeName": "connection_id",
                                    "KeyType": "RANGE",
                                },
                            ],
                        }
                    ],
                },
            }
        resource_definitions = {"Resources": resources}
        return resource_definitions

    def build_role_statement_yml(self) -> ExternalItemType:
        role_statements = [
            {
                "Effect": "Allow",
                "Action": [
                    "dynamodb:BatchWriteItem",
                    "dynamodb:DeleteItem",
                    "dynamodb:GetItem",
                    "dynamodb:PutItem",
                    "dynamodb:Query",
                ],
                "Resource": [
                    {
                        "Fn::GetAtt": [
                            self.connection_table_name.title().replace("_", ""),
                            "Arn",
                        ]
                    },
                    {
                        "Fn::Join": [
                            "/",
                            [
                                {
                                    "Fn::GetAtt": [
                                        self.connection_table_name.title().replace(
                                            "_", ""
                                        ),
                                        "Arn",
                                    ]
                                },
                                "index",
                                "gsi__subscription_name__connection_id",
                            ],
                        ]
                    },
                ],
            },
            {
                "Effect": "Allow",
                "Action": [
                    "sqs:SendMessage",
                    "sqs:ReceiveMessage",
                    "sqs:DeleteMessage",
                    "sqs:GetQueueUrl",
                ],
                "Resource": [
                    {"Fn::GetAtt": [s.name.title().replace("_", "") + "SQS", "Arn"]}
                    for s in self.subscriptions
                    if s.action_subscribers
                ],
            },
        ]
        subscription_policy = {"iamRoleStatements": role_statements}
        return subscription_policy



from dataclasses import dataclass
from typing import Optional


@dataclass
class ErrorResponse:
    """Content definition for error response"""

    error_code: str
    error_msg: Optional[str] = None



from abc import ABC

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC, T


class SingletonABC(ABC):
    """Abstract singleton implementation"""

    def __new__(cls):
        instance = getattr(cls, "__instance", None)
        if instance is None:
            instance = object.__new__(cls)
            setattr(cls, "__instance", instance)
        return instance

    def __repr__(self):
        return self.__class__.__name__

    # noinspection PyUnusedLocal
    @classmethod
    def __marshaller_factory__(cls, marshaller_context):
        return SingletonMarshaller(cls)


class SingletonMarshaller(MarshallerABC[T]):
    def load(self, item: ExternalType) -> T:
        return self.marshalled_type()

    def dump(self, item: T) -> ExternalType:
        return self.marshalled_type.__name__



from abc import ABC, abstractmethod

from marshy.types import ExternalItemType

from servey.action.action import Action
from servey.trigger.trigger_abc import TriggerABC


class TriggerHandlerABC(ABC):
    @abstractmethod
    def handle_trigger(
        self,
        action: Action,
        trigger: TriggerABC,
        lambda_definition: ExternalItemType,
    ):
        """Add definitions for the trigger given to the lambda_definition"""



import base64
import os
from logging import getLogger

from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import AuthorizerFactoryABC

LOGGER = getLogger(__name__)


class JwtAuthorizerFactory(AuthorizerFactoryABC):
    """
    Factory for authorizers that try to generate / validate jwt tokens using a private key
    defined specified as an environment variable: JWT_SECRET_KEY
    """

    priority = 50

    def create_authorizer(self) -> AuthorizerABC:
        try:
            from servey.security.authorizer.jwt_authorizer import JwtAuthorizer

            jwt_secret_key = os.environ.get("JWT_SECRET_KEY")
            if jwt_secret_key is None:
                LOGGER.warning(
                    "JWT_SECRET_KEY NOT DEFINED - Making One Up! (Auth will fail between restarts!)"
                )
                jwt_secret_key = base64.b64encode(os.urandom(12)).decode("UTF-8")
            authorizer = JwtAuthorizer(jwt_secret_key)
            return authorizer
        except ModuleNotFoundError:
            LOGGER.info("PyJWT is not available - skipping JwtAuthorizer")



import inspect
from collections import namedtuple
from unittest import TestCase

from starlette.requests import Request
from strawberry.types import Info

from servey.action.action import action, get_action
from servey.security.access_control.scope_access_control import ScopeAccessControl
from servey.security.authorization import Authorization, ROOT, AuthorizationError
from servey.servey_strawberry.handler_filter.authorization_handler_filter import (
    AuthorizationHandlerFilter,
)
from servey.servey_strawberry.schema_factory import SchemaFactory


class TestAuthorizationHandlerFilter(TestCase):
    def filtered_action(self):
        # noinspection PyUnusedLocal
        @action(access_control=ScopeAccessControl("root"))
        def dummy(title: str, auth: Authorization) -> str:
            """Dummy"""

        action_ = get_action(dummy)
        filter_ = AuthorizationHandlerFilter()
        schema_factory = SchemaFactory()
        filtered_action, continue_filtering = filter_.filter(action_, schema_factory)
        self.assertTrue(continue_filtering)

        sig = inspect.signature(filtered_action.fn)
        params = list(sig.parameters.values())
        expected_params = [
            inspect.Parameter(
                "title", inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=str
            ),
            inspect.Parameter("info", inspect.Parameter.KEYWORD_ONLY, annotation=Info),
        ]
        self.assertEqual(expected_params, params)
        return filtered_action, filter_

    def test_filter(self):
        filtered_action, filter_ = self.filtered_action()
        token = filter_.authorizer.encode(ROOT)
        request = Request(
            dict(
                method="GET",
                type="http",
                headers=[
                    [
                        "authorization".encode("latin-1"),
                        f"Bearer {token}".encode("latin-1"),
                    ]
                ],
            )
        )
        context = dict(request=request)
        raw_info = namedtuple("RawInfo", ["context"])(context=context)
        # noinspection PyTypeChecker
        info = Info(raw_info, None)
        filtered_action.fn(title="foobar", info=info)
        filtered_action.fn(title="foobar", info=info)

    def test_filter_no_auth(self):
        filtered_action, filter_ = self.filtered_action()
        request = Request(dict(method="GET", type="http", headers=[]))
        context = dict(request=request)
        raw_info = namedtuple("RawInfo", ["context"])(context=context)
        # noinspection PyTypeChecker
        info = Info(raw_info, None)
        with self.assertRaises(AuthorizationError):
            filtered_action.fn(title="foobar", info=info)



from datetime import datetime

from servey.action.action import action
from servey.trigger.web_trigger import WEB_GET


@action(triggers=(WEB_GET,))
def get_the_time() -> datetime:
    """This is a dummy"""



import os
from unittest import TestCase

from servey.finder.module_subscription_finder import ModuleSubscriptionFinder


class TestModuleSubscriptionFinder(TestCase):
    def test_default_constructor(self):
        finder = ModuleSubscriptionFinder()
        self.assertEqual(ModuleSubscriptionFinder("servey_main.subscriptions"), finder)
        os.environ["SERVEY_MAIN"] = "foobar"
        finder = ModuleSubscriptionFinder()
        self.assertEqual(ModuleSubscriptionFinder("foobar.subscriptions"), finder)
        os.environ.pop("SERVEY_MAIN")

    def test_find_subscriptions(self):
        from tests.finder.subscriptions.test_subscriptions import my_subscription

        finder = ModuleSubscriptionFinder("tests.finder.subscriptions")
        subscriptions = list(finder.find_subscriptions())
        expected = [my_subscription]
        self.assertEqual(expected, subscriptions)

    def test_find_subscriptions_not_found(self):
        finder = ModuleSubscriptionFinder("this.module.does.not.exist")
        self.assertEqual([], list(finder.find_subscriptions()))



from unittest import TestCase

from servey.action.action import action, get_action, Action
from servey.trigger.fixed_rate_trigger import FixedRateTrigger
from servey.trigger.web_trigger import WEB_GET
from servey.cache_control.ttl_cache_control import TtlCacheControl
from servey.action.example import Example
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.access_control.scope_access_control import ScopeAccessControl


class TestAction(TestCase):
    def test_decorator_no_args(self):
        # noinspection PyUnusedLocal
        @action
        def noop(value: int) -> str:
            """Noop"""

        action_ = get_action(noop)
        expected = Action(
            fn=noop,
            name="noop",
            description="Noop",
            access_control=ALLOW_ALL,
            triggers=tuple(),
            timeout=15,
            examples=None,
            cache_control=None,
        )
        self.assertEqual(expected, action_)

    def test_decorator_args(self):
        # noinspection PyUnusedLocal
        @action(
            access_control=ScopeAccessControl("execute"),
            triggers=(WEB_GET, FixedRateTrigger(3600)),
            timeout=30,
            examples=(Example(name="ten", params=dict(value=10), result="That's 10"),),
            cache_control=TtlCacheControl(900),
        )
        def noop(value: int) -> str:
            """Noop"""

        action_ = get_action(noop)
        expected = Action(
            fn=noop,
            name="noop",
            description="Noop",
            access_control=ScopeAccessControl("execute"),
            triggers=(WEB_GET, FixedRateTrigger(3600)),
            timeout=30,
            examples=(Example(name="ten", params=dict(value=10), result="That's 10"),),
            cache_control=TtlCacheControl(900),
        )
        self.assertEqual(expected, action_)



import base64
import json
from base64 import urlsafe_b64encode
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Optional, Dict

import boto3
import jwt
from jwt import DecodeError
from schemey.util import filter_none

from servey.security.authorization import Authorization, AuthorizationError
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.jwt_authorizer import date_from_jwt


@dataclass
class KmsAuthorizer(AuthorizerABC):
    """
    Authorizer that uses kms to work with tokens, offloading the responsibility for key
    management to AWS. I would have loved to use cognito for this, but was hampered
    by it not having the ability to specify custom per user scopes in the access token.

    Encoding a JWT token requires use of the private key, and always goes out to kms.
    Verification only uses the public key, which is cached locally.

    Note: When decoding, this implementation only grabs what is available in KMS
    it makes no distinction about filtering to a particular subset of keys. That sort
    of logic would probably require a class with an attached persistence mechanism
    """

    key_id: str
    iss: Optional[str] = None
    aud: Optional[str] = None
    kms: Optional[Any] = None
    public_keys: Optional[Dict[str, str]] = field(default_factory=dict)

    def __post_init__(self):
        if not self.kms:
            self.kms = boto3.client("kms")

    def encode(self, authorization: Authorization) -> str:
        header = urlsafe_b64encode(
            json.dumps(dict(typ="JWT", alg="RS256", kid=self.key_id)).encode()
        )
        payload = urlsafe_b64encode(
            json.dumps(
                filter_none(
                    dict(
                        iss=self.iss,
                        sub=authorization.subject_id,
                        aud=self.aud,
                        exp=int(authorization.expire_at.timestamp())
                        if authorization.expire_at
                        else None,
                        nbf=int(authorization.not_before.timestamp())
                        if authorization.not_before
                        else None,
                        iat=int(datetime.now().timestamp()),
                        scope=" ".join(authorization.scopes),
                    )
                )
            ).encode()
        )
        message = header + b"." + payload
        # noinspection SpellCheckingInspection
        result = self.kms.sign(
            Message=message,
            KeyId=self.key_id,
            SigningAlgorithm="RSASSA_PKCS1_V1_5_SHA_256",
            MessageType="RAW",
        )
        signature = urlsafe_b64encode(result["Signature"])
        result = message + b"." + signature
        return result.decode("UTF-8")

    @staticmethod
    def get_key_id(token: str) -> str:
        """May need the key id from the token header in order to figure out which key to use"""
        headers = jwt.get_unverified_header(token)
        return headers["kid"]

    def get_public_key(self, key_id: str):
        public_key = self.public_keys.get(key_id)
        if not public_key:
            response = self.kms.get_public_key(KeyId=key_id)
            public_key = response["PublicKey"]
            public_key = base64.b64encode(public_key).decode()
            public_key = (
                "-----BEGIN PUBLIC KEY-----\n"
                + public_key
                + "\n-----END PUBLIC KEY-----"
            ).encode()
            # noinspection PyTypeChecker
            self.public_keys[public_key] = public_key

        return public_key

    def authorize(self, token: str) -> Authorization:
        """
        Verifying tokens only requires the public key. We cache this from kms to do verifications
        locally which is more efficient
        """
        try:
            header = jwt.get_unverified_header(token)
            key_id = header.get("kid")
            public_key = self.get_public_key(key_id)
            decoded = jwt.decode(jwt=token, key=public_key, algorithms=["RS256"])
            authorization = Authorization(
                subject_id=decoded.get("sub"),
                not_before=date_from_jwt(decoded, "nbf"),
                expire_at=date_from_jwt(decoded, "exp"),
                scopes=frozenset(decoded.get("scope").split(" ")),
            )
            return authorization
        except DecodeError as e:
            raise AuthorizationError(e)



import json
from dataclasses import dataclass, field
from typing import Optional, Dict, List, Any

import boto3

from servey.servey_aws import is_lambda_env
from servey.subscription.subscription import Subscription, T
from servey.subscription.subscription_service import (
    SubscriptionServiceABC,
    SubscriptionServiceFactoryABC,
)
from servey.util import get_servey_main


@dataclass
class SqsSubscriptionService(SubscriptionServiceABC):
    """Subscription service backed by SQS."""

    queue_names_by_subscription_name: Dict[str, str]
    sqs_client: Any = field(default_factory=lambda: boto3.client("sqs"))
    queue_urls_by_queue_name: Dict[str, Optional[str]] = field(default_factory=dict)

    def get_queue_url(self, subscription_name: str) -> Optional[str]:
        queue_name = self.queue_names_by_subscription_name.get(subscription_name)
        if not queue_name:
            return
        queue_url = self.queue_urls_by_queue_name.get(queue_name)
        if not queue_url:
            response = self.sqs_client.get_queue_url(QueueName=queue_name)
            queue_url = self.queue_urls_by_queue_name[queue_name] = response["QueueUrl"]
        return queue_url

    def publish(self, subscription: Subscription[T], event: T):
        queue_url = self.get_queue_url(subscription.name)
        if queue_url:
            self.sqs_client.send_message(
                QueueUrl=queue_url,
                MessageBody=json.dumps(subscription.event_marshaller.dump(event)),
            )


class SqsSubscriptionServiceFactory(SubscriptionServiceFactoryABC):
    def create(
        self, subscriptions: List[Subscription]
    ) -> Optional[SubscriptionServiceABC]:
        if not is_lambda_env():
            return
        queue_names_by_subscription_name = {}
        service_name = get_servey_main()
        for subscription in subscriptions:
            if subscription.action_subscribers:
                queue_name = service_name + "-" + subscription.name
                queue_names_by_subscription_name[subscription.name] = queue_name
        if queue_names_by_subscription_name:
            return SqsSubscriptionService(queue_names_by_subscription_name)



import inspect
from dataclasses import dataclass
from datetime import datetime

from marshy.types import ExternalItemType

from servey.finder.action_finder_abc import find_actions_with_trigger_type, find_actions
from servey.trigger.web_trigger import WebTrigger, WebTriggerMethod
from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    ensure_ref_in_file,
    GENERATED_HEADER,
    create_yml_file,
)
from servey.util import get_servey_main


@dataclass
class AppsyncConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    servey_actions_yml_file: str = "serverless_servey/appsync.yml"
    servey_appsync_schema_file: str = "serverless_servey/schema.graphql"

    def configure(self, main_serverless_yml_file: str):
        ensure_ref_in_file(
            main_serverless_yml_file,
            ["custom", "appSync"],
            self.servey_actions_yml_file,
        )
        self.build_appsync_schema()
        servey_action_functions_yml = self.build_servey_action_functions_yml()
        create_yml_file(self.servey_actions_yml_file, servey_action_functions_yml)

    def build_appsync_schema(self):
        from servey.servey_strawberry.schema_factory import create_schema

        with open(self.servey_appsync_schema_file, "w") as writer:
            writer.write("# ")
            writer.write(GENERATED_HEADER.replace("\n", "\n# "))
            writer.write(f"\n# Updated at: {datetime.now().isoformat()}\n\n")
            schema = create_schema()

            schema = str(schema)

            # Using string manipulation here feels really gross, but I have not been able to
            # get directives working in strawberry...

            # noinspection SpellCheckingInspection
            schema = schema.replace(
                '\n"""Date with time (isoformat)"""\nscalar DateTime\n', ""
            )
            schema = schema.replace("DateTime", "AWSDateTime")
            writer.write(schema)

    def build_servey_action_functions_yml(self) -> ExternalItemType:
        appsync_definitions = {
            "name": get_servey_main(),
            "authenticationType": "API_KEY",
            "defaultMappingTemplates": {"request": False, "response": False},
            "mappingTemplates": [],
            "dataSources": [],
            "schema": self.servey_appsync_schema_file,
            # caching:
            #    behavior: FULL_REQUEST_CACHING  # or PER_RESOLVER_CACHING. Required
            #    ttl: 3600  # The TTL of the cache. Optional. Default: 3600
            #    atRestEncryption:  # Bool, Optional. Enable at rest encryption. disabled by default.
            #    transitEncryption:  # Bool, Optional. Enable transit encryption. disabled by default.
            #    type: 'T2_SMALL'  # Cache instance size. Optional. Default: 'T2_SMALL'
        }
        for action, trigger in find_actions_with_trigger_type(WebTrigger):
            field = action.name.title().replace("_", "")
            mapping_template = {
                "dataSource": action.name,
                "type": "Query"
                if trigger.method == WebTriggerMethod.GET
                else "Mutation",
                "field": field[0].lower() + field[1:],
                # caching:
                #    keys:  # array. A list of VTL variables to use as cache key.
                #        - '$context.identity.sub'
                #        - '$context.arguments.id'
                #    ttl: 1000  # override the ttl for this resolver. (default comes from global config)
            }
            appsync_definitions["mappingTemplates"].append(mapping_template)
            data_source = {
                "name": action.name,
                "type": "AWS_LAMBDA",
                "config": {
                    "functionName": action.name,
                },
            }
            if action.description:
                data_source["description"] = action.description.strip()
            appsync_definitions["dataSources"].append(data_source)

        for action in find_actions():
            sig = inspect.signature(action.fn)
            params = list(sig.parameters.values())
            if params and params[0].name == "self":
                type_name, field_name = action.fn.__qualname__.split(".")
                field_name = field_name[0] + field_name.title()[1:].replace("_", "")
                mapping_template = {
                    "dataSource": action.name,
                    "type": type_name,
                    "field": field_name
                    # caching:
                    #    keys:  # array. A list of VTL variables to use as cache key.
                    #        - '$context.identity.sub'
                    #        - '$context.arguments.id'
                    #    ttl: 1000  # override the ttl for this resolver. (default comes from global config)
                }
                appsync_definitions["mappingTemplates"].append(mapping_template)
                # noinspection PyTypeChecker
                data_source = next(
                    (
                        d
                        for d in appsync_definitions["dataSources"]
                        if d["name"] == action.name
                    ),
                    None,
                )
                if not data_source:
                    data_source = {
                        "name": action.name,
                        "type": "AWS_LAMBDA",
                        "config": {
                            "functionName": action.name,
                        },
                    }
                    if action.description:
                        data_source["description"] = action.description.strip()
                    appsync_definitions["dataSources"].append(data_source)

        return appsync_definitions



from abc import abstractmethod, ABC
from typing import Type, Optional

SchemaFactory = "servey.servey_strawberry.schema_factory.SchemaFactory"


class EntityFactoryABC(ABC):
    priority: int = 100

    @abstractmethod
    def create_type(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        """Create a type"""

    @abstractmethod
    def create_input(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        """Create an input"""



from typing import Type, Optional

import typing_inspect
from marshy.utils import resolve_forward_refs

from servey.servey_strawberry.entity_factory.entity_factory_abc import (
    EntityFactoryABC,
)
from servey.servey_strawberry.schema_factory import SchemaFactory


class ForwardRefFactory(EntityFactoryABC):
    priority: int = 200

    def create_type(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        if typing_inspect.is_forward_ref(annotation):
            annotation = resolve_forward_refs(annotation)
            return schema_factory.get_type(annotation)

    def create_input(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        if typing_inspect.is_forward_ref(annotation):
            annotation = resolve_forward_refs(annotation)
            return schema_factory.get_input(annotation)



from enum import Enum, EnumMeta
from unittest import TestCase

from servey.servey_strawberry.entity_factory.enum_factory import EnumFactory
from servey.servey_strawberry.schema_factory import SchemaFactory


class TestEnumFactory(TestCase):
    def test_create_input(self):
        input_ = EnumFactory().create_input(_FooBar, SchemaFactory())
        self.assertIs(EnumMeta, input_.__class__)
        # noinspection PyTypeChecker
        self.assertEqual(["foo", "bar"], [e.value for e in input_])


class _FooBar(Enum):
    FOO = "foo"
    BAR = "bar"



import importlib
import json
import os
from unittest import TestCase
from unittest.mock import patch, MagicMock

from servey.security.authorization import ROOT
from servey.subscription.subscription import subscription
from tests.servey_strawberry.test_schema_factory import NumberStats


class TestLambdaWebsocket(TestCase):
    def test_connect(self):
        mock_resource = MagicMock()
        with (
            patch("boto3.resource", mock_resource),
            patch.dict(
                os.environ,
                {
                    "AWS_REGION": "bezos-moon-base",
                    "CONNECTION_TABLE_NAME": "mock_connections",
                },
            ),
        ):
            from servey.servey_aws import lambda_websocket

            importlib.reload(lambda_websocket)
            lambda_websocket.lambda_handler(
                {
                    "requestContext": {
                        "connectionId": "connection_1",
                        "routeKey": "$connect",
                        "domainName": "foobar.com",
                        "stage": "dev",
                    },
                    "headers": {
                        "Authorization": f"Bearer {lambda_websocket._AUTHORIZER.encode(ROOT)}"
                    },
                },
                None,
            )
            lambda_websocket.lambda_handler(
                {
                    "requestContext": {
                        "connectionId": "connection_2",
                        "routeKey": "$connect",
                        "domainName": "foobar2.com",
                        "stage": "dev",
                    }
                },
                None,
            )

            table = mock_resource.return_value.Table
            table.assert_called_with("mock_connections")
            put_item = table.return_value.put_item
            put_item_call_list = [a.kwargs["Item"] for a in put_item.call_args_list]
            expected = [
                {
                    "connection_id": "connection_1",
                    "subscription_name": " ",
                    "user_authorization": {
                        "subject_id": None,
                        "scopes": ["root"],
                        "not_before": None,
                        "expire_at": None,
                    },
                    "updated_at": put_item_call_list[0]["updated_at"],
                },
                {
                    "connection_id": "connection_2",
                    "subscription_name": " ",
                    "user_authorization": None,
                    "updated_at": put_item_call_list[1]["updated_at"],
                },
            ]
            self.assertEqual(expected, put_item_call_list)

    def test_disconnect(self):
        mock_resource = MagicMock()
        items = [
            {"connection_id": "connection_1", "subscription_name": " "},
            {"connection_id": "connection_1", "subscription_name": "foo"},
            {"connection_id": "connection_1", "subscription_name": "bar"},
        ]
        mock_resource.return_value.Table.return_value.query.return_value = {
            "Items": items,
        }
        with (
            patch("boto3.resource", mock_resource),
            patch.dict(
                os.environ,
                {
                    "AWS_REGION": "bezos-moon-base",
                    "CONNECTION_TABLE_NAME": "mock_connections",
                },
            ),
        ):
            from servey.servey_aws import lambda_websocket

            importlib.reload(lambda_websocket)
            result = lambda_websocket.lambda_handler(
                {
                    "requestContext": {
                        "connectionId": "connection_1",
                        "routeKey": "$disconnect",
                        "domainName": "foobar.com",
                        "stage": "dev",
                    },
                },
                None,
            )
            self.assertEqual({"statusCode": 200}, result)
            table = mock_resource.return_value.Table.return_value
            delete_item = (
                table.batch_writer.return_value.__enter__.return_value.delete_item
            )
            delete_item_call_list = [
                a.kwargs["Key"] for a in delete_item.call_args_list
            ]
            self.assertEqual(items, delete_item_call_list)

    def test_subscribe(self):
        subscription_ = subscription(NumberStats, "number_stats")
        mock_resource = MagicMock()
        item = {"connection_id": "connection_1", "subscription_name": " "}
        table = mock_resource.return_value.Table.return_value
        table.get_item.return_value = item
        with (
            patch("boto3.resource", mock_resource),
            patch.dict(
                os.environ,
                {
                    "AWS_REGION": "bezos-moon-base",
                    "CONNECTION_TABLE_NAME": "mock_connections",
                },
            ),
            patch(
                "servey.finder.subscription_finder_abc.find_subscriptions",
                return_value=[subscription_],
            ),
        ):
            from servey.servey_aws import lambda_websocket

            importlib.reload(lambda_websocket)
            result = lambda_websocket.lambda_handler(
                {
                    "requestContext": {
                        "connectionId": "connection_1",
                        "routeKey": "$default",
                        "domainName": "foobar.com",
                        "stage": "dev",
                    },
                    "body": json.dumps(
                        {"type": "Subscribe", "payload": "number_stats"}
                    ),
                },
                None,
            )
            self.assertEqual({"statusCode": 200}, result)
            put_item = mock_resource.return_value.Table.return_value.put_item
            put_item_call_list = [a.kwargs["Item"] for a in put_item.call_args_list]
            self.assertEqual(
                [
                    {
                        "connection_id": "connection_1",
                        "subscription_name": "number_stats",
                        "endpoint_url": "https://foobar.com/dev",
                        "updated_at": put_item_call_list[0]["updated_at"],
                        "user_authorization": None,
                    }
                ],
                put_item_call_list,
            )

    def test_unsubscribe(self):
        subscription_ = subscription(NumberStats, "number_stats")
        mock_resource = MagicMock()
        with (
            patch("boto3.resource", mock_resource),
            patch.dict(
                os.environ,
                {
                    "AWS_REGION": "bezos-moon-base",
                    "CONNECTION_TABLE_NAME": "mock_connections",
                },
            ),
            patch(
                "servey.finder.subscription_finder_abc.find_subscriptions",
                return_value=[subscription_],
            ),
        ):
            from servey.servey_aws import lambda_websocket

            importlib.reload(lambda_websocket)
            result = lambda_websocket.lambda_handler(
                {
                    "requestContext": {
                        "connectionId": "connection_1",
                        "routeKey": "$default",
                        "domainName": "foobar.com",
                        "stage": "dev",
                    },
                    "body": json.dumps(
                        {"type": "Unsubscribe", "payload": "number_stats"}
                    ),
                },
                None,
            )
            self.assertEqual({"statusCode": 200}, result)
            delete_item = mock_resource.return_value.Table.return_value.delete_item
            delete_item_call_list = [
                a.kwargs["Key"] for a in delete_item.call_args_list
            ]
            self.assertEqual(
                [
                    {
                        "connection_id": "connection_1",
                        "subscription_name": "number_stats",
                    }
                ],
                delete_item_call_list,
            )

    def test_invalid_body(self):
        subscription_ = subscription(NumberStats, "number_stats")
        mock_resource = MagicMock()
        with (
            patch("boto3.resource", mock_resource),
            patch.dict(
                os.environ,
                {
                    "AWS_REGION": "bezos-moon-base",
                    "CONNECTION_TABLE_NAME": "mock_connections",
                },
            ),
            patch(
                "servey.finder.subscription_finder_abc.find_subscriptions",
                return_value=[subscription_],
            ),
        ):
            from servey.servey_aws import lambda_websocket

            importlib.reload(lambda_websocket)
            response = lambda_websocket.lambda_handler(
                {
                    "requestContext": {
                        "connectionId": "connection_1",
                        "routeKey": "$default",
                        "domainName": "foobar.com",
                        "stage": "dev",
                    },
                    "body": "This is not valid",
                },
                None,
            )
            self.assertEqual({"statusCode": 400}, response)



from typing import List, Union

from marshy.types import ExternalItemType
from schemey import Schema

from servey.action.util import move_ref_items_to_components
from servey.security.access_control.allow_none import ALLOW_NONE
from servey.subscription.subscription import Subscription


def subscription_event_schema(
    subscriptions: List[Subscription], components: ExternalItemType
) -> Schema:
    schemas = [s.event_schema for s in subscriptions if s.access_control != ALLOW_NONE]
    any_of = []
    for subscription in subscriptions:
        schema = subscription.event_schema.schema
        schema = move_ref_items_to_components(schema, schema, components)
        any_of.append(
            {
                "properties": {"type": {"const": type.__name__}, "payload": schema},
                "additionalProperties": False,
                "required": ["type", "payload"],
            }
        )
    type_ = Union[tuple(s.python_type for s in schemas)]
    return Schema({"anyOf": any_of}, type_)



import os
from unittest import TestCase
from unittest.mock import patch

from servey.servey_strawberry.strawberry_starlette_route_factory import (
    StrawberryStarletteRouteFactory,
)


class TestStrawberryStarletteRouteFactory(TestCase):
    def test_create_routes(self):
        with patch.dict(
            os.environ,
            {"SERVEY_MAIN": "tests.servey_strawberry"},
        ):
            factory = StrawberryStarletteRouteFactory()
            routes = list(factory.create_routes())
            self.assertEqual(3, len(routes))

    def test_no_debug(self):
        with patch.dict(
            os.environ,
            {
                "SERVEY_MAIN": "tests.servey_strawberry",
                "SERVER_DEBUG": "0",
            },
        ):
            factory = StrawberryStarletteRouteFactory()
            routes = list(factory.create_routes())
            self.assertEqual(2, len(routes))

    def test_create_no_routes(self):
        with patch("servey.finder.action_finder_abc.find_actions", return_value=[]):
            factory = StrawberryStarletteRouteFactory()
            routes = list(factory.create_routes())
            self.assertEqual([], routes)

    def test_create_routes_no_module(self):
        def raise_module_not_found():
            raise ModuleNotFoundError()

        factory = StrawberryStarletteRouteFactory()
        with patch(
            "servey.servey_strawberry.schema_factory.create_schema",
            raise_module_not_found,
        ):
            routes = list(factory.create_routes())
            self.assertEqual([], routes)



from datetime import datetime
from unittest import TestCase

from servey.cache_control.ttl_cache_control import TtlCacheControl


class TestTtlCacheControl(TestCase):
    def test_get_cache_header(self):
        cache_control = TtlCacheControl(900)
        cache_header = cache_control.get_cache_header(dict(values=[1, "str", True]))
        self.assertIsNotNone(cache_header.etag)
        self.assertAlmostEqual(
            cache_header.expire_at.timestamp(),
            datetime.now().timestamp() + 900,
            delta=1,
        )



from unittest import TestCase

from servey.security.access_control.allow_all import AllowAll, ALLOW_ALL
from servey.security.authorization import PUBLIC


class TestAllowAll(TestCase):
    def test_singleton(self):
        self.assertIs(ALLOW_ALL, AllowAll())

    def test_methods(self):
        self.assertTrue(ALLOW_ALL.is_executable(PUBLIC))



from marshy.types import ExternalItemType

from servey.action.action import Action
from servey.errors import ServeyError
from servey.trigger.fixed_rate_trigger import FixedRateTrigger
from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
    TriggerHandlerABC,
)
from servey.trigger.trigger_abc import TriggerABC

UNITS = {"days": 86400, "hours": 3600, "minutes": 60}


class FixedRateTriggerHandler(TriggerHandlerABC):
    def handle_trigger(
        self,
        action: Action,
        trigger: TriggerABC,
        lambda_definition: ExternalItemType,
    ):
        if not isinstance(trigger, FixedRateTrigger):
            return
        events = lambda_definition.get("events")
        if not events:
            events = lambda_definition["events"] = []
        if trigger.interval < 60:
            raise ServeyError("frequency_too_high")  # min is 60!
        for unit, seconds in UNITS.items():
            if not trigger.interval % seconds:
                # noinspection PyTypeChecker
                events.append(
                    dict(
                        schedule=dict(
                            rate=f"rate({int(trigger.interval / seconds)} {unit})"
                        )
                    )
                )
                return



from unittest import TestCase
from unittest.mock import patch

from servey.action.action import action, get_action
from servey.action.example import Example
from servey.servey_test.test_servey_actions import define_test_class


class TestTestServeyActions(TestCase):
    def test_define_test_class(self):
        with patch(
            "servey.servey_test.test_servey_actions.find_actions",
            return_value=[get_action(say_hello)],
        ):
            # noinspection PyPep8Naming
            TestClass = define_test_class()
            self.assertEqual("TestServeyActions", TestClass.__name__)
            self.assertTrue(issubclass(TestClass, TestCase))
            keys = {k for k in TestClass.__dict__.keys() if not k.startswith("__")}
            self.assertEqual({"test_say_hello__world", "test_say_hello__bad"}, keys)
            test_instance = TestClass()
            test_instance.test_say_hello__world()
            with self.assertRaises(AssertionError):
                # This example should fail!
                test_instance.test_say_hello__bad()


@action(
    examples=(
        Example("world", params=dict(name="World"), result="Hello World!"),
        Example("bad", params=dict(name="Wrong"), result="Nope"),
    )
)
def say_hello(name: str) -> str:
    return f"Hello {name}!"



from datetime import datetime
from unittest import TestCase

from dateutil.relativedelta import relativedelta

from servey.security.authorization import (
    Authorization,
    AuthorizationError,
    ROOT,
    get_inject_at,
)


class TestAuthorization(TestCase):
    def test_is_valid_for_timestamp(self):
        authorization = Authorization(
            None,
            frozenset("root"),
            datetime.fromisoformat("2022-01-01"),
            datetime.fromisoformat("2022-02-01"),
        )
        self.assertTrue(
            authorization.is_valid_for_timestamp(datetime.fromisoformat("2022-01-02"))
        )
        self.assertTrue(
            authorization.is_valid_for_timestamp(datetime.fromisoformat("2022-01-01"))
        )
        self.assertTrue(
            authorization.is_valid_for_timestamp(datetime.fromisoformat("2022-02-01"))
        )
        self.assertFalse(
            authorization.is_valid_for_timestamp(
                datetime.fromisoformat("2022-02-01 00:00:01")
            )
        )
        self.assertFalse(
            authorization.is_valid_for_timestamp(
                datetime.fromisoformat("2021-12-31 23:59:59")
            )
        )

    def test_is_valid_for_timestamp_no_min(self):
        authorization = Authorization(
            None, frozenset("root"), None, datetime.fromisoformat("2022-02-01")
        )
        self.assertTrue(
            authorization.is_valid_for_timestamp(datetime.fromisoformat("2022-01-01"))
        )
        self.assertTrue(
            authorization.is_valid_for_timestamp(datetime.fromisoformat("2022-02-01"))
        )
        self.assertFalse(
            authorization.is_valid_for_timestamp(
                datetime.fromisoformat("2022-02-01 00:00:01")
            )
        )

    def test_is_valid_for_timestamp_no_max(self):
        authorization = Authorization(
            None, frozenset("root"), datetime.fromisoformat("2022-02-01"), None
        )
        self.assertFalse(
            authorization.is_valid_for_timestamp(datetime.fromisoformat("2022-01-01"))
        )
        self.assertTrue(
            authorization.is_valid_for_timestamp(datetime.fromisoformat("2022-02-01"))
        )
        self.assertTrue(
            authorization.is_valid_for_timestamp(
                datetime.fromisoformat("2022-02-01 00:00:01")
            )
        )

    def test_is_valid_for_timestamp_no_timestamp(self):
        authorization = Authorization(None, frozenset("root"), None, None)
        self.assertTrue(
            authorization.is_valid_for_timestamp(datetime.fromisoformat("2022-02-01"))
        )

    def test_is_valid_for_timestamp_now(self):
        a_day = relativedelta(days=1)
        authorization = Authorization(
            None,
            frozenset("root"),
            datetime.now() - a_day,
            datetime.now() + a_day,
        )
        self.assertTrue(authorization.is_valid_for_timestamp())

    def test_has_scope(self):
        authorization = Authorization(None, frozenset(["foo", "bar"]), None, None)
        self.assertTrue(authorization.has_scope("foo"))
        self.assertFalse(authorization.has_scope("zap"))

    def test_has_any_scope(self):
        authorization = Authorization(None, frozenset(["foo", "bar"]), None, None)
        self.assertTrue(authorization.has_any_scope({"foo"}))
        self.assertFalse(authorization.has_any_scope({"zap"}))
        self.assertTrue(authorization.has_any_scope({"foo", "zap"}))

    def test_has_all_scope(self):
        authorization = Authorization(None, frozenset(["foo", "bar"]), None, None)
        self.assertTrue(authorization.has_all_scopes({"foo"}))
        self.assertTrue(authorization.has_all_scopes({"foo", "bar"}))
        self.assertFalse(authorization.has_all_scopes({"zap"}))
        self.assertFalse(authorization.has_all_scopes({"foo", "zap"}))

    def test_check_valid_for_timestamp(self):
        authorization = Authorization(
            None,
            frozenset("root"),
            datetime.fromisoformat("2022-01-01"),
            datetime.fromisoformat("2022-02-01"),
        )
        authorization.check_valid_for_timestamp(datetime.fromisoformat("2022-01-02"))
        authorization.check_valid_for_timestamp(datetime.fromisoformat("2022-01-01"))
        authorization.check_valid_for_timestamp(datetime.fromisoformat("2022-02-01"))
        with self.assertRaises(AuthorizationError):
            authorization.check_valid_for_timestamp(
                datetime.fromisoformat("2022-02-01 00:00:01")
            )
        with self.assertRaises(AuthorizationError):
            authorization.check_valid_for_timestamp(
                datetime.fromisoformat("2021-12-31 23:59:59")
            )

    def test_check_scope(self):
        authorization = Authorization(None, frozenset(["foo", "bar"]), None, None)
        authorization.check_scope("foo")
        with self.assertRaises(AuthorizationError):
            self.assertFalse(authorization.check_scope("zap"))

    def test_check_any_scope(self):
        authorization = Authorization(None, frozenset(["foo", "bar"]), None, None)
        authorization.check_any_scope({"foo"})
        with self.assertRaises(AuthorizationError):
            authorization.check_any_scope({"zap"})
        authorization.check_any_scope({"foo", "zap"})

    def test_check_all_scope(self):
        authorization = Authorization(None, frozenset(["foo", "bar"]), None, None)
        authorization.check_all_scopes({"foo"})
        authorization.check_all_scopes({"foo", "bar"})
        with self.assertRaises(AuthorizationError):
            authorization.check_all_scopes({"foo", "zap"})

    def test_get_inject_at(self):
        # noinspection PyUnusedLocal
        def foo(bar: str, auth: Authorization) -> str:
            return f"bar:{bar}, auth:{auth}"

        foo("bar", ROOT)  # Nonsense
        self.assertEqual("auth", get_inject_at(foo))



from abc import abstractmethod, ABC

from servey.security.authorization import Authorization


class AuthorizerABC(ABC):
    @abstractmethod
    def authorize(self, token: str) -> Authorization:
        """Convert the token given to an authorization object"""

    @abstractmethod
    def encode(self, authorization: Authorization) -> str:
        """Encode the authorization given as a string which may be included in an authorization header"""



import asyncio
import json
import os
from unittest import TestCase
from unittest.mock import patch

from servey.action.action import action, get_action
from servey.security.access_control.scope_access_control import ScopeAccessControl
from servey.security.authorization import Authorization, ROOT
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_starlette.route_factory.action_route_factory import (
    ActionRouteFactory,
)
from servey.servey_starlette.route_factory.openapi_route_factory import (
    OpenapiRouteFactory,
)
from servey.trigger.web_trigger import WEB_GET
from tests.servey_starlette.action_endpoint.test_action_endpoint import build_request


class TestAuthorizingActionEndpoint(TestCase):
    def test_valid_input_get(self):
        @action(triggers=(WEB_GET,), access_control=ScopeAccessControl("root"))
        def echo_get(val: str, auth: Authorization) -> str:
            return " ".join(auth.scopes) + ": " + val

        with (
            patch(
                "servey.servey_starlette.route_factory.action_route_factory.find_actions",
                return_value=[get_action(echo_get)],
            ),
            patch(
                "servey.servey_starlette.route_factory.openapi_route_factory.find_actions",
                return_value=[get_action(echo_get)],
            ),
            patch.dict(os.environ, {"SERVER_HOST": "https://testy.com"}),
        ):
            routes = list(ActionRouteFactory().create_routes())
            openapi_routes = list(OpenapiRouteFactory().create_routes())
            self.assertEqual(1, len(routes))
            self.assertEqual(2, len(openapi_routes))
            loop = asyncio.get_event_loop()

            # Unauthorized
            request = build_request(query_string="val=bar")
            response = loop.run_until_complete(routes[0].endpoint(request))
            self.assertEqual(401, response.status_code)
            self.assertEqual(dict(error="unauthorized"), json.loads(response.body))

            # Authorized
            authorizer = get_default_authorizer()
            token = authorizer.encode(ROOT)
            request = build_request(
                query_string="val=bar",
                headers=dict(Authorization=f"Bearer {token}"),
            )
            response = loop.run_until_complete(routes[0].endpoint(request))
            self.assertEqual(200, response.status_code)
            self.assertEqual("root: bar", json.loads(response.body))

            request = build_request()
            response = openapi_routes[0].endpoint(request)
            self.assertEqual(200, response.status_code)
            schema = json.loads(response.body)
            expected_schema = {
                "components": {
                    "ErrorResponse": {
                        "additionalProperties": False,
                        "description": "Content definition for error response",
                        "name": "ErrorResponse",
                        "properties": {
                            "error_code": {"type": "string"},
                            "error_msg": {
                                "anyOf": [{"type": "string"}, {"type": "null"}],
                                "default": None,
                            },
                        },
                        "required": ["error_code"],
                        "type": "object",
                    },
                    "securitySchemes": {
                        "OAuth2PasswordBearer": {
                            "flows": {"password": {"scopes": {}, "tokenUrl": "/login"}},
                            "type": "oauth2",
                        }
                    },
                },
                "info": {"title": "Servey", "version": "0.1.0"},
                "openapi": "3.0.2",
                "paths": {
                    "/actions/echo-get": {
                        "get": {
                            "operationId": "echo_get",
                            "parameters": [
                                {
                                    "in": "query",
                                    "name": "val",
                                    "required": True,
                                    "schema": {"type": "string"},
                                }
                            ],
                            "responses": {
                                "200": {
                                    "content": {
                                        "application/json": {
                                            "schema": {"type": "string"}
                                        }
                                    },
                                    "description": "Successful Response",
                                },
                                "403": {
                                    "content": {
                                        "application/json": {
                                            "schema": {
                                                "$ref": "#components/ErrorResponse"
                                            }
                                        }
                                    },
                                    "description": "unauthorized",
                                },
                                "422": {"description": "Validation " "Error"},
                            },
                            "security": [{"OAuth2PasswordBearer": []}],
                        }
                    }
                },
                "servers": [{"url": "https://testy.com"}],
            }
            self.assertEqual(expected_schema, schema)



import os
from dataclasses import replace
from unittest import TestCase

from servey.action.action import action
from servey.finder.module_action_finder import ModuleActionFinder


class TestModuleActionFinder(TestCase):
    def test_default_constructor(self):
        finder = ModuleActionFinder()
        self.assertEqual(ModuleActionFinder("servey_main.actions"), finder)
        os.environ["SERVEY_MAIN"] = "foobar"
        finder = ModuleActionFinder()
        self.assertEqual(ModuleActionFinder("foobar.actions"), finder)
        os.environ.pop("SERVEY_MAIN")

    def test_find_actions(self):
        finder = ModuleActionFinder("tests.finder.test_module_action_finder")
        actions = list(finder.find_actions())
        expected = [
            replace(foo.__servey_action__, fn=actions[0].fn),
            replace(bar.__servey_action__, fn=actions[1].fn),
        ]
        self.assertEqual(expected, actions)
        self.assertIsNone(actions[0].fn())
        self.assertEqual("Value: 5", actions[1].fn(5))

    def test_find_actions_not_found(self):
        finder = ModuleActionFinder("this.module.does.not.exist")
        self.assertEqual([], list(finder.find_actions()))


@action
def foo() -> type(None):
    pass


@action
def bar(value: int) -> str:
    return f"Value: {value}"


foo()
bar(10)



from unittest import TestCase

from servey.action.action import action, get_action
from servey.errors import ServeyError
from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory import (
    ActionEndpointFactory,
)
from servey.trigger.web_trigger import WebTrigger, WebTriggerMethod, WEB_POST


class TestActionEndpointFactory(TestCase):
    def test_multi_paths(self):
        # noinspection PyTypeChecker
        @action(
            triggers=(
                WebTrigger(WebTriggerMethod.POST, "/foo"),
                WebTrigger(WebTriggerMethod.POST, "/bar"),
            )
        )
        def dummy() -> int:
            """No action required"""

        factory = ActionEndpointFactory()
        with self.assertRaises(ServeyError):
            factory.create(get_action(dummy), set(), [])

    def test_custom_path(self):
        # noinspection PyTypeChecker
        @action(triggers=(WebTrigger(WebTriggerMethod.GET, "/foo"),))
        def dummy() -> int:
            """No action required"""

        factory = ActionEndpointFactory()
        endpoint = factory.create(get_action(dummy), set(), [])
        route = endpoint.get_route()
        # noinspection PyTypeChecker
        self.assertEqual("/foo", route.path)

    def test_invalid_return(self):
        # noinspection PyTypeChecker
        @action(triggers=(WEB_POST,))
        def dummy():
            """No action required"""

        factory = ActionEndpointFactory()
        with self.assertRaises(ServeyError):
            factory.create(get_action(dummy), set(), [])



import logging
from dataclasses import dataclass, field, fields, is_dataclass
from typing import Type, Dict, List, Set

import strawberry
import typing_inspect
from marshy.factory.impl_marshaller_factory import get_impls
from marshy.factory.optional_marshaller_factory import get_optional_type
from strawberry.annotation import StrawberryAnnotation

# noinspection PyProtectedMember
from strawberry.field import StrawberryField, UNRESOLVED
from strawberry.type import StrawberryOptional, StrawberryContainer
from strawberry.types.fields.resolver import StrawberryResolver

from servey.action.action import Action
from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.trigger.web_trigger import WebTrigger, UPDATE_METHODS
from servey.servey_strawberry.entity_factory.entity_factory_abc import (
    EntityFactoryABC,
)
from servey.servey_strawberry.handler_filter.handler_filter_abc import (
    HandlerFilterABC,
)

LOGGER = logging.getLogger(__name__)


@dataclass
class SchemaFactory:
    """
    This class is tasked with converting standard python types into types usable by strawberry and therefore graphql.
    """

    types: Dict[str, Type] = field(default_factory=dict)
    inputs: Dict[str, Type] = field(default_factory=dict)
    enums: Dict[str, Type] = field(default_factory=dict)
    query: Dict[str, StrawberryField] = field(default_factory=dict)
    mutation: Dict[str, StrawberryField] = field(default_factory=dict)
    subscription: Dict[str, StrawberryField] = field(default_factory=dict)
    entity_factories: List[EntityFactoryABC] = field(default_factory=list)
    handler_filters: List[HandlerFilterABC] = field(default_factory=list)

    def get_input(self, annotation: Type) -> Type:
        if hasattr(annotation, "__name__"):
            i = self.inputs.get(annotation.__name__)
            if i:
                return i
        for entity_factory in self.entity_factories:
            i = entity_factory.create_input(annotation, self)
            if i:
                if hasattr(annotation, "__name__"):
                    self.inputs[annotation.__name__] = i
                return i

    def get_type(self, annotation: Type):
        if hasattr(annotation, "__name__"):
            type_ = self.types.get(annotation.__name__)
            if type_:
                return type_
        for entity_factory in self.entity_factories:
            type_ = entity_factory.create_type(annotation, self)
            if type_:
                if hasattr(annotation, "__name__"):
                    self.types[annotation.__name__] = type_
                return type_

    def create_field_for_action(self, action: Action, trigger: WebTrigger):
        for handler_filter in self.handler_filters:
            action, continue_filtering = handler_filter.filter(action, self)
            if not continue_filtering:
                break

        f = strawberry.field(resolver=action.fn)
        f.name = action.name
        if trigger.method in UPDATE_METHODS:
            self.mutation[f.name] = f
        else:
            self.query[f.name] = f

    def _resolve_type_futures(self, type_, resolved: Set):
        if isinstance(type_, str):
            type_ = self.types[type_]
        if isinstance(type_, StrawberryAnnotation):
            type_.type = self._resolve_type_futures(type_.annotation, resolved)
            return type_
        if isinstance(type_, StrawberryContainer):
            type_.of_type = self._resolve_type_futures(type_.of_type, resolved)
            return type_
        name = typing_inspect.get_forward_arg(type_)
        if name:
            type_ = self.types[name]
            return type_
        optional_type = get_optional_type(type_)
        if optional_type:
            return StrawberryOptional(
                self._resolve_type_futures(optional_type, resolved)
            )
        origin = typing_inspect.get_origin(type_)
        if origin:
            args = tuple(
                self._resolve_type_futures(a, resolved)
                for a in typing_inspect.get_args(type_)
            )
            if origin is list:
                return List[args]
            else:
                return origin[args]
        if is_dataclass(type_):
            if type_.__name__ in resolved:
                return type_
            resolved.add(type_.__name__)
            # noinspection PyDataclass
            for f in fields(type_):
                if isinstance(f, StrawberryField):
                    if f.type is UNRESOLVED:
                        resolver = f.base_resolver
                        field_type = resolver.signature.return_annotation
                        field_type = self._resolve_type_futures(field_type, resolved)
                        resolver_override = StrawberryResolver(
                            resolver.wrapped_func, type_override=field_type
                        )
                        f.base_resolver = resolver_override
                else:
                    f.type = self._resolve_type_futures(f.type, resolved)
        return type_

    def create_schema(self):
        resolved = set()
        for f in self.query.values():
            f.type = self._resolve_type_futures(f.type, resolved)
        for f in self.mutation.values():
            f.type = self._resolve_type_futures(f.type, resolved)

        query_params = {
            **self.query,
            "__annotations__": {f.name: f.type for f in self.query.values()},
        }
        queries = (
            strawberry.type(type("Query", (), query_params)) if self.query else None
        )

        mutation_params = {
            **self.mutation,
            "__annotations__": {f.name: f.type for f in self.mutation.values()},
        }
        mutations = (
            strawberry.type(type("Mutation", (), mutation_params))
            if self.mutation
            else None
        )

        subscription_params = {
            **self.subscription,
            "__annotations__": {s.name: s.type for s in self.subscription.values()},
        }
        subscriptions = (
            strawberry.type(type("Subscription", (), subscription_params))
            if self.subscription
            else None
        )

        if not queries:
            LOGGER.warning("No graphql queries found - skipping schema generation")
            return None  # strawberry requires an API has at least 1 query!
        schema = strawberry.Schema(queries, mutations, subscriptions)
        return schema


def create_schema_factory() -> SchemaFactory:
    entity_factories = [f() for f in get_impls(EntityFactoryABC)]
    entity_factories.sort(key=lambda f: f.priority, reverse=True)
    handler_filters = [f() for f in get_impls(HandlerFilterABC)]
    handler_filters.sort(key=lambda f: f.priority, reverse=True)
    schema_factory = SchemaFactory(
        entity_factories=entity_factories, handler_filters=handler_filters
    )
    return schema_factory


def create_schema():
    schema_factory = create_schema_factory()
    for action, trigger in find_actions_with_trigger_type(WebTrigger):
        schema_factory.create_field_for_action(action, trigger)
    schema = schema_factory.create_schema()
    return schema



import asyncio
import json
import logging
from dataclasses import dataclass, field
from typing import Iterator, Optional, Dict, List, Set
from uuid import uuid4

from starlette.endpoints import WebSocketEndpoint
from starlette.routing import Route, WebSocketRoute
from starlette.websockets import WebSocket

from servey.errors import ServeyError
from servey.finder.subscription_finder_abc import find_subscriptions
from servey.security.access_control.allow_none import ALLOW_NONE
from servey.security.authorization import Authorization
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC
from servey.subscription.subscription import Subscription, T
from servey.subscription.subscription_service import (
    SubscriptionServiceFactoryABC,
    SubscriptionServiceABC,
)

LOGGER = logging.getLogger(__name__)


@dataclass
class SubscriptionRouteFactory(RouteFactoryABC, SubscriptionServiceFactoryABC):
    path: str = "/subscription"

    def create_routes(self) -> Iterator[Route]:
        if _get_subscription_connections_by_name():
            route = WebSocketRoute(self.path, _SubscriptionEndpoint)
            yield route

    def create(
        self, subscriptions: List[Subscription]
    ) -> Optional[SubscriptionServiceABC]:
        if _get_subscription_connections_by_name():
            return _StarletteSubscriptionService()


class _SubscriptionEndpoint(WebSocketEndpoint):
    async def on_connect(self, websocket: WebSocket) -> None:
        await websocket.accept()
        authorization = None
        token = websocket.headers.get("Authorization")
        if token and token.startswith("Bearer "):
            token = token[7:]
            authorizer = get_default_authorizer()
            authorization = authorizer.authorize(token)

        connection = _Connection(
            connection_id=str(uuid4()), websocket=websocket, authorization=authorization
        )
        LOGGER.debug("connect:{connection_id}", connection_id=connection.connection_id)
        websocket.path_params["connection_id"] = connection.connection_id
        _get_connections_by_id()[connection.connection_id] = connection

    async def on_disconnect(self, websocket: WebSocket, close_code: int) -> None:
        connection_id = websocket.path_params["connection_id"]
        LOGGER.debug("disconnect:{connection_id}", connection_id=connection_id)
        connections_by_id = _get_connections_by_id()
        connection = connections_by_id.pop(connection_id)
        subscription_connections_by_name = _get_subscription_connections_by_name()
        for subscription_name in connection.subscription_names:
            subscription_connections_by_name[subscription_name].connections.remove(
                connection
            )

    async def on_receive(self, websocket: WebSocket, data: str) -> None:
        event = json.loads(data)
        type_ = event["type"]
        connection_id = websocket.path_params["connection_id"]
        connection = _get_connections_by_id().get(connection_id)
        if not connection:
            LOGGER.warning(f"unknown_connection:{connection_id}")
            return
        if type_ == "Subscribe":
            subscription_name = event["payload"]
            if subscription_name in connection.subscription_names:
                return
            subscription_connection = _get_subscription_connections_by_name()[
                subscription_name
            ]
            if not subscription_connection.subscription.access_control.is_executable(
                connection.authorization
            ):
                raise ServeyError("unauthorized")
            subscription_connection.connections.append(connection)
            connection.subscription_names.add(subscription_name)
        elif type_ == "Unsubscribe":
            subscription_name = event["payload"]
            if subscription_name not in connection.subscription_names:
                return
            subscription_connection = _get_subscription_connections_by_name()[
                subscription_name
            ]
            subscription_connection.connections.remove(connection)
            connection.subscription_names.remove(subscription_name)
        else:
            raise ServeyError(f"unknown_type:{type_}")


@dataclass
class _Connection:
    connection_id: str
    websocket: WebSocket
    subscription_names: Set[str] = field(default_factory=set)
    authorization: Optional[Authorization] = None


@dataclass
class _SubscriptionConnections:
    subscription: Subscription
    connections: List[_Connection] = field(default_factory=list)


_SUBSCRIPTION_CONNECTIONS_BY_NAME: Optional[Dict[str, _SubscriptionConnections]] = None
_CONNECTIONS_BY_ID: Dict[str, _Connection] = {}


def _get_subscription_connections_by_name() -> Dict[str, _SubscriptionConnections]:
    global _SUBSCRIPTION_CONNECTIONS_BY_NAME
    if _SUBSCRIPTION_CONNECTIONS_BY_NAME is None:
        _SUBSCRIPTION_CONNECTIONS_BY_NAME = {
            s.name: _SubscriptionConnections(s)
            for s in find_subscriptions()
            if s.access_control != ALLOW_NONE
        }
    return _SUBSCRIPTION_CONNECTIONS_BY_NAME


def _get_connections_by_id() -> Dict[str, _Connection]:
    return _CONNECTIONS_BY_ID


class _StarletteSubscriptionService(SubscriptionServiceABC):
    def publish(self, subscription: Subscription[T], event: T):
        subscription_connections = _get_subscription_connections_by_name().get(
            subscription.name
        )
        if not subscription_connections:
            return
        loop = asyncio.get_event_loop()
        json_event = subscription.event_marshaller.dump(event)
        for connection in subscription_connections.connections:
            # noinspection PyBroadException
            try:
                if (
                    subscription.event_filter
                    and not subscription.event_filter.should_publish(
                        event, connection.authorization
                    )
                ):
                    continue
                sending = connection.websocket.send_json(json_event)
                loop.create_task(sending)
            except Exception:
                LOGGER.exception("failed_to_send_data")



from unittest import TestCase

from servey.cache_control.cache_header import CacheHeader
from servey.cache_control.secure_hash_cache_control import SecureHashCacheControl


class TestSecureHashCacheControl(TestCase):
    def test_get_cache_header(self):
        cache_control = SecureHashCacheControl()
        cache_header = cache_control.get_cache_header(dict(values=[1, "str", True]))
        # noinspection SpellCheckingInspection
        expected = CacheHeader("IjLwXxgJicusWH98zrk+2VzPUZ+FuDMknuo3fdOzIEs=")
        self.assertEqual(expected, cache_header)

    def test_get_cache_header_public(self):
        cache_control = SecureHashCacheControl(private=False)
        cache_header = cache_control.get_cache_header(dict(values=[2, "str", True]))
        # noinspection SpellCheckingInspection
        expected = CacheHeader(
            "HxYzJNnJVRCdsnux0p3vdUajMtJFiZY1FFKXsF0RUPc=", private=False
        )
        self.assertEqual(expected, cache_header)



from dataclasses import dataclass
from typing import Generic, TypeVar, Optional, Type, Tuple

from marshy import get_default_context
from marshy.marshaller.marshaller_abc import MarshallerABC
from schemey import Schema, schema_from_type

from servey.action.action import Action
from servey.security.access_control.access_control_abc import AccessControlABC
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.subscription.event_filter_abc import EventFilterABC

T = TypeVar("T")


@dataclass(frozen=True)
class Subscription(Generic[T]):
    name: str
    event_marshaller: MarshallerABC[T]
    event_schema: Schema
    access_control: AccessControlABC = ALLOW_ALL
    event_filter: Optional[EventFilterABC] = None
    action_subscribers: Tuple[Action, ...] = tuple()

    def publish(self, event: T):
        """
        Publish an event to subscribers
        """
        from servey.subscription.subscription_service import get_subscription_services

        for subscription_service in get_subscription_services():
            subscription_service.publish(self, event)


def subscription(
    event_type: Type[T],
    name: Optional[str] = None,
    access_control: AccessControlABC = ALLOW_ALL,
    event_filter: Optional[EventFilterABC] = None,
    action_subscribers: Tuple[Action] = tuple(),
) -> Subscription[T]:
    subscription_ = Subscription(
        event_marshaller=get_default_context().get_marshaller(event_type),
        event_schema=schema_from_type(event_type),
        name=name or event_type.__name__,
        access_control=access_control,
        event_filter=event_filter,
        action_subscribers=action_subscribers,
    )
    return subscription_



import dataclasses
import inspect
from dataclasses import dataclass, field
from inspect import Parameter
from typing import Optional, Tuple

from marshy.factory.optional_marshaller_factory import get_optional_type
from strawberry.types import Info

from servey.action.action import Action
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.authorization import Authorization, AuthorizationError
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_strawberry.handler_filter.handler_filter_abc import (
    HandlerFilterABC,
)
from servey.servey_strawberry.schema_factory import SchemaFactory


@dataclass
class AuthorizationHandlerFilter(HandlerFilterABC):
    priority: int = 120
    authorizer: AuthorizerABC = field(default_factory=get_default_authorizer)
    # What to name the authorization parameter if none exists.
    info_kwarg_name: str = "info"  # NOTE: It seems that strawberry reserves the name: Info for this parameter

    def filter(
        self,
        action: Action,
        schema_factory: SchemaFactory,
    ) -> Tuple[Action, bool]:
        fn = action.fn
        sig = inspect.signature(fn)

        parameters = []
        authorization_kwarg_name = None
        for param in sig.parameters.values():
            type_ = get_optional_type(param.annotation) or param.annotation
            if type_ is Authorization:
                authorization_kwarg_name = param.name
            else:
                parameters.append(param)

        if not authorization_kwarg_name and action.access_control == ALLOW_ALL:
            return action, True

        if authorization_kwarg_name:
            parameters.append(
                Parameter(
                    name=self.info_kwarg_name,
                    kind=Parameter.KEYWORD_ONLY,
                    annotation=Info,
                )
            )

        def resolver(**kwargs):
            info = kwargs.pop(self.info_kwarg_name)
            authorization = self.get_authorization(info)
            if not action.access_control.is_executable(authorization):
                raise AuthorizationError()
            if authorization_kwarg_name:
                kwargs[authorization_kwarg_name] = authorization
            result = fn(**kwargs)
            return result

        sig = sig.replace(parameters=parameters)
        resolver.__signature__ = sig
        wrapped_action = dataclasses.replace(action, fn=resolver)
        return wrapped_action, True

    def get_authorization(self, info: Info) -> Optional[Authorization]:
        authorization = info.context.get("authorization")
        if authorization:
            return authorization
        request = info.context.get("request")
        if request:
            token = request.headers.get("Authorization")
            if token and token.lower().startswith("bearer "):
                token = token[7:]
                authorization = self.authorizer.authorize(token)
                info.context["authorization"] = authorization
                return authorization



from servey.security.access_control.access_control_abc import (
    AccessControlABC,
)
from servey.security.authorization import Authorization
from servey.util.singleton_abc import SingletonABC


class AllowAll(SingletonABC, AccessControlABC):
    def is_executable(self, authorization: Authorization) -> bool:
        return True


ALLOW_ALL = AllowAll()



import json
import os
from unittest import TestCase
from unittest.mock import patch

from marshy.marshaller import NoOpMarshaller
from schemey import schema_from_type

from servey.servey_starlette.route_factory.asyncapi_route_factory import (
    AsyncapiRouteFactory,
)
from servey.subscription.subscription import Subscription
from tests.servey_starlette.action_endpoint.test_action_endpoint import build_request


class TestAsyncapiRouteFactory(TestCase):
    def test_endpoint(self):
        with (
            patch(
                "servey.servey_starlette.route_factory.asyncapi_route_factory.find_subscriptions",
                return_value=[
                    Subscription(
                        "my_message", NoOpMarshaller(str), schema_from_type(str)
                    ),
                ],
            ),
            patch.dict(os.environ, {"SERVER_WS_URL": "https://foo.com/"}),
        ):
            route = AsyncapiRouteFactory()
            response = route.endpoint(build_request())
            result = json.loads(response.body)
            expected_result = {
                "asyncapi": "2.5.0",
                "info": {"title": "Servey", "version": "0.1.0"},
                "channels": {
                    "/subscription": {
                        "publish": {
                            "message": {
                                "payload": {
                                    "properties": {
                                        "type": {"enum": ["Subscribe", "Unsubscribe"]},
                                        "payload": {"type": "string"},
                                    },
                                    "additionalProperties": False,
                                }
                            }
                        },
                        "subscribe": {
                            "message": {
                                "payload": {
                                    "anyOf": [
                                        {
                                            "properties": {
                                                "type": {"const": "type"},
                                                "payload": {"type": "string"},
                                            },
                                            "additionalProperties": False,
                                            "required": ["type", "payload"],
                                        }
                                    ]
                                }
                            }
                        },
                    }
                },
                "components": {},
                "servers": {
                    "production": {"url": "https://foo.com/", "protocol": "wss"}
                },
            }
            self.assertEqual(expected_result, result)



from servey.action.action import get_action
from servey.security.access_control.allow_none import ALLOW_NONE
from servey.security.access_control.scope_access_control import ScopeAccessControl
from servey.subscription.subscription import subscription
from tests.specs.number_spec.actions import integer_stats_consumer
from tests.specs.number_spec.models import IntegerStats


integer_stats_for_admins = subscription(
    event_type=IntegerStats,
    name="integer_stats_for_admins",
    access_control=ScopeAccessControl("root"),
)


integer_stats_queue = subscription(
    event_type=IntegerStats,
    name="integer_stats_queue",
    access_control=ALLOW_NONE,
    action_subscribers=(get_action(integer_stats_consumer),),
)



import importlib
import json
import os
from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, Callable, List
from unittest import TestCase
from unittest.mock import patch

from marshy import dump

from servey.action.action import action
from servey.action.batch_invoker import BatchInvoker
from servey.cache_control.timestamp_cache_control import TimestampCacheControl
from servey.errors import ServeyError
from servey.security.access_control.scope_access_control import ScopeAccessControl
from servey.security.authorization import Authorization, ROOT, AuthorizationError
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.trigger.web_trigger import WEB_POST, WEB_GET
from tests.servey_strawberry.test_schema_factory import Node


class TestLambdaInvoker(TestCase):
    def test_direct(self):
        invoker = get_invoker(get_node)
        event = dict(params=dict(name="foo"))
        result = invoker(event, None)
        expected_result = {"child_nodes": [], "name": "foo"}
        self.assertEqual(expected_result, result)

    def test_direct_with_auth(self):
        invoker = get_invoker(get_authorization)
        event = dict(params={})
        with self.assertRaises(AuthorizationError):
            invoker(event, None)

        authorizer = get_default_authorizer()
        event = dict(params={}, authorization=authorizer.encode(ROOT))
        result = invoker(event, None)
        expected_result = {
            "subject_id": None,
            "scopes": ["root"],
            "not_before": None,
            "expire_at": None,
        }
        self.assertEqual(expected_result, result)

        event = dict(params={}, authorization=dump(ROOT))
        result = invoker(event, None)
        expected_result = {
            "subject_id": None,
            "scopes": ["root"],
            "not_before": None,
            "expire_at": None,
        }
        self.assertEqual(expected_result, result)

    def test_appsync_get(self):
        invoker = get_invoker(get_node)
        event = dict(arguments=dict(name="foo"))
        result = invoker(event, None)
        expected_result = {"child_nodes": [], "name": "foo"}
        self.assertEqual(expected_result, result)

    def test_appsync_post(self):
        invoker = get_invoker(put_node)
        event = dict(arguments=dict(node=dict(name="bar")))
        result = invoker(event, None)
        self.assertEqual(True, result)
        self.assertEqual(Node("bar"), _NODES["bar"])

    def test_api_gateway_get(self):
        invoker = get_invoker(get_node)
        event = dict(httpMethod="GET", queryStringParameters=dict(name="foo"))
        result = invoker(event, None)
        # noinspection PyTypeChecker
        expected_result = {
            "statusCode": 200,
            "headers": {
                "Content-Type": "application/json",
            },
            "body": '{"name": "foo", "child_nodes": []}',
        }
        self.assertEqual(expected_result, result)

    def test_api_gateway_post(self):
        invoker = get_invoker(put_node)
        event = dict(httpMethod="POST", body=json.dumps(dict(node=dict(name="foo"))))
        result = invoker(event, None)
        expected_result = {
            "statusCode": 200,
            "headers": {"Content-Type": "application/json"},
            "body": "true",
        }
        self.assertEqual(expected_result, result)

    def test_api_gateway_with_auth(self):
        invoker = get_invoker(get_authorization)
        event = dict(httpMethod="POST", body="{}")
        with self.assertRaises(AuthorizationError):
            invoker(event, None)

        authorizer = get_default_authorizer()
        event = dict(
            httpMethod="POST",
            body="{}",
            headers={"Authorization": f"Bearer {authorizer.encode(ROOT)}"},
        )
        result = invoker(event, None)
        expected_result = {
            "statusCode": 200,
            "headers": {"Content-Type": "application/json"},
            "body": json.dumps(dump(ROOT)),
        }
        self.assertEqual(expected_result, result)

    def test_api_gateway_with_caching(self):
        invoker = get_invoker(get_stored_item)
        event = dict(httpMethod="GET", queryStringParameters=dict(name="foo"))
        result = invoker(event, None)
        # noinspection PyTypeChecker,SpellCheckingInspection
        headers = {
            "Content-Type": "application/json",
            "ETag": "3g6+tMQd4iiW50nY1WFSohCR/S4DGhUbtq20jRcr5so=",
            "Cache-Control": "no-storage",
            "Last-Modified": "Wed, 01 Jan 2020 07:00:00 GMT",
        }
        expected_result = {
            "statusCode": 200,
            "headers": headers,
            "body": '{"name": "foo", "updated_at": "2020-01-01T07:00:00+00:00"}',
        }
        self.assertEqual(expected_result, result)
        # noinspection SpellCheckingInspection
        event["headers"] = {
            "If-Match": "3g6+tMQd4iiW50nY1WFSohCR/S4DGhUbtq20jRcr5so=",
        }
        result = invoker(event, None)
        # noinspection PyTypeChecker
        expected_result = {
            "statusCode": 304,
            "headers": headers,
            "body": "",
        }
        self.assertEqual(expected_result, result)
        # noinspection PyTypeChecker
        event["headers"] = {
            "If-Modified-Since": "Thu, 02 Jan 2020 07:00:00 GMT",
        }
        result = invoker(event, None)
        # noinspection PyTypeChecker,SpellCheckingInspection
        expected_result = {
            "statusCode": 304,
            "headers": {
                "Content-Type": "application/json",
                "Cache-Control": "no-storage",
                "ETag": "3g6+tMQd4iiW50nY1WFSohCR/S4DGhUbtq20jRcr5so=",
                "Last-Modified": result["headers"]["Last-Modified"],
            },
            "body": "",
        }
        self.assertEqual(expected_result, result)

    def test_appsync_post_nested(self):
        with patch.dict(
            os.environ,
            {
                "SERVEY_ACTION_MODULE": __name__,
                "SERVEY_ACTION_FUNCTION_NAME": "Node.tree_size",
            },
        ):
            from servey.servey_aws import lambda_invoker

            importlib.reload(lambda_invoker)
            result = lambda_invoker.invoke(
                {"arguments": {}, "source": {"name": "foobar"}}, None
            )
            self.assertEqual(1, result)

    def test_appsync_post_auth(self):
        authorizer = get_default_authorizer()
        token = authorizer.encode(ROOT)
        invoker = get_invoker(get_authorization)
        event = {
            "arguments": {},
            "request": {"headers": {"authorization": f"Bearer {token}"}},
        }
        result = invoker(event, None)
        self.assertEqual(
            {
                "subject_id": None,
                "scopes": ["root"],
                "not_before": None,
                "expire_at": None,
            },
            result,
        )

        event = {
            "arguments": {},
        }
        with self.assertRaises(AuthorizationError):
            invoker(event, None)

    def test_invalid(self):
        invoker = get_invoker(get_node)
        event = {}
        with self.assertRaises(ServeyError):
            invoker(event, None)

    def test_nested(self):
        with patch.dict(
            os.environ,
            {
                "SERVEY_ACTION_MODULE": __name__,
                "SERVEY_ACTION_FUNCTION_NAME": "Node.tree_size",
            },
        ):
            from servey.servey_aws import lambda_invoker

            importlib.reload(lambda_invoker)
            result = lambda_invoker.invoke(
                {"params": {"self": {"name": "foobar"}}}, None
            )
            self.assertEqual(1, result)

    def test_sqs(self):
        invoker = get_invoker(double)
        event = {
            "Records": [
                {"body": "3"},
                {"body": "5"},
            ]
        }
        result = invoker(event, None)
        self.assertEqual([6, 10], result)

    def test_sqs_batched(self):
        invoker = get_invoker(double_batched)
        event = {
            "Records": [
                {"body": "3"},
                {"body": "5"},
            ]
        }
        result = invoker(event, None)
        self.assertEqual([6, 10], result)

    def test_sqs_invalid(self):
        invoker = get_invoker(add)
        event = {
            "Records": [
                {"body": "[1,2]"},
                {"body": "[3,4]"},
            ]
        }
        with self.assertRaises(ServeyError):
            invoker(event, None)


_NODES = {"foo": Node("foo")}


@action(triggers=(WEB_GET,))
def get_node(name: str) -> Optional[Node]:
    return _NODES.get(name)


@action(triggers=(WEB_POST,))
def put_node(node: Node) -> bool:
    _NODES[node.name] = node
    return True


@action(triggers=(WEB_POST,), access_control=ScopeAccessControl("root"))
async def get_authorization(
    authorization: Optional[Authorization],
) -> Optional[Authorization]:
    return authorization


def get_invoker(fn: Callable):
    # noinspection PyUnresolvedReferences
    with patch.dict(
        os.environ,
        {
            "SERVEY_ACTION_MODULE": fn.__module__,
            "SERVEY_ACTION_FUNCTION_NAME": fn.__name__,
        },
    ):
        from servey.servey_aws import lambda_invoker

        importlib.reload(lambda_invoker)
        return lambda_invoker.invoke


@dataclass
class StoredItem:
    name: str
    updated_at: datetime = field(default_factory=datetime.now)


# noinspection PyUnusedLocal
@action(triggers=(WEB_GET,), cache_control=TimestampCacheControl())
def get_stored_item(name: str) -> Optional[StoredItem]:
    return StoredItem("foo", datetime.fromisoformat("2020-01-01"))


@action
def double(value: int) -> int:
    return value * 2


def batched_double(values: List[int]) -> List[int]:
    results = [i * 2 for i in values]
    return results


# noinspection PyUnusedLocal
@action(batch_invoker=BatchInvoker(fn=batched_double))
def double_batched(value: int) -> int:
    """Never actually called!"""


# noinspection PyUnusedLocal
@action
def add(a: int, b: int) -> int:
    """Never actually called!"""



"""
Allows invoking actions directly from the command line (Or via crontab)
Usage:

`python -m servey.servey_direct --action=ping`
`python -m servey.servey_direct --action=say_hello "--event={\"name\": \"Foobar\"}"`

"""
import argparse
import inspect
import json

from marshy import get_default_context

from servey.action.util import get_marshaller_for_params
from servey.finder.action_finder_abc import find_actions

parser = argparse.ArgumentParser(description="Invoke an action directly")
parser.add_argument("--run", default="action")
parser.add_argument("--action")
parser.add_argument("--event", default="{}")
args = parser.parse_args()


action = next((a for a in find_actions() if a.name == args.action), None)
if not action:
    raise ValueError(f"no_such_action:{args.action}")
marshaller = get_marshaller_for_params(action.fn, set())
kwargs = marshaller.load(json.loads(args.event))
result = action.fn(**kwargs)
return_annotation = inspect.signature(action.fn).return_annotation
if return_annotation != inspect.Parameter.empty:
    result = get_default_context().dump(result, return_annotation)
    result_str = json.dumps(result)
    print(result_str)



from datetime import datetime, timezone

from marshy.marshaller import DatetimeMarshaller


class ToSecondDatetimeMarshaller(DatetimeMarshaller):
    """
    Marshaller which does not dump microseconds - appropriate for server instances where this precision is
    not useful (And Appsync scalars don't support it)
    """

    def dump(self, item: datetime) -> str:
        dumped = item.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S+00:00")
        return dumped
